import abc
import functools
import io
import importlib
import os
import shelve
import shutil
from pathlib import Path
from typing import Any, Callable, Dict, Generator, FrozenSet, Generic, IO, Iterable, Iterator, \
    Mapping, MutableMapping, MutableSequence, Optional, Set, Tuple, TypeVar, Union

from frozendict import frozendict


def _immutable(value: Any) -> Any:
    """
    Creates a read-only version from the specified value.

    Dictionaries, lists, and sets will be converted recursively.

    :param value: Value to be transformed into a read-only version
    :return: Read-only value
    """
    if isinstance(value, dict):
        return frozendict({k: _immutable(v) for k, v in value.items()})
    elif isinstance(value, set):
        return frozenset({_immutable(v) for v in value})
    elif isinstance(value, list):
        return tuple([_immutable(v) for v in value])
    else:
        return value


def _mutable(value: Any) -> Any:
    """
    Creates a writeable version from the specified (read-only) value.

    Frozen dictionaries, tuples, and frozen sets will be converted recursively.

    :param value: Value to be transformed into a writeable version
    :return: Writeable value
    """
    if isinstance(value, frozendict):
        return {k: _mutable(v) for k, v in value.items()}
    elif isinstance(value, frozenset):
        return {_mutable(v) for v in value}
    elif isinstance(value, tuple):
        return [_mutable(v) for v in value]
    else:
        return value


class Asset:
    """
    Represents a digital asset.

    An `Asset` is an immutable value object whose contents consist of *essence*
    and *metadata*. Essence represents the actual data of a media file, such as
    the color values of an image, whereas the metadata describes the essence.

    Assets should not be instantiated directly. Instead, use
    :func:`~madam.core.Madam.read` to retrieve an `Asset` representing the
    content.
    """
    def __init__(self, essence: IO, **metadata: Any) -> None:
        """
        Initializes a new `Asset` with the specified essence and metadata.

        :param essence: The essence of the asset as a file-like object
        :type essence: IO
        :param \\**metadata: The metadata describing the essence
        :type \\*metadata: Any
        """
        self._essence_data = essence.read()
        if 'mime_type' not in metadata:
            metadata['mime_type'] = None
        self.metadata = _immutable(metadata)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Asset):
            return NotImplemented
        return other.__dict__ == self.__dict__

    def __getattr__(self, item: str) -> Any:
        if item in self.metadata:
            return self.metadata[item]
        raise AttributeError(f'{self.__class__!r} object has no attribute {item!r}')

    def __setattr__(self, key: str, value: Any):
        if 'metadata' in self.__dict__ and key in self.__dict__['metadata']:
            raise NotImplementedError('Unable to overwrite metadata attribute.')
        super().__setattr__(key, value)

    def __setstate__(self, state: Dict[str, Any]):
        """
        Sets this objects __dict__ to the specified state.

        Required for Asset to be unpicklable. If this is absent, pickle will
        not set the `__dict__` correctly due to the presence of
        :func:`~madam.core.Asset.__getattr__`.

        :param state: The state passed by pickle
        """
        self.__dict__ = state

    @property
    def essence(self) -> IO:
        """
        Represents the actual content of the asset.

        The essence of an MP3 file, for example, is only comprised of the actual audio data,
        whereas metadata such as ID3 tags are stored separately as metadata.
        """
        return io.BytesIO(self._essence_data)

    def __hash__(self) -> int:
        return hash(self._essence_data) ^ hash(self.metadata)

    def __repr__(self) -> str:
        metadata_str = ' '.join(
            '{}={!r}'.format(k, v)
            for k, v in self.metadata.items()
            if not isinstance(v, frozendict)
        )
        return '<{} {}>'.format(self.__class__.__qualname__, metadata_str)


class UnsupportedFormatError(Exception):
    """
    Represents an error that is raised whenever file content with unknown type
    is encountered.
    """
    def __init__(self, *args) -> None:
        """
        Initializes a new `UnsupportedFormatError`.
        """
        super().__init__(*args)


def operator(function: Callable[..., Asset]) -> Callable[..., Callable[..., Asset]]:
    """
    Decorator function for methods that process assets.

    Usually, it will be used with operations in a :class:`~madam.core.Processor`
    implementation to make the methods configurable before applying the method
    to an asset.

    Only keyword arguments are allowed for configuration.

    Example for using a decorated :attr:`convert` method:

    .. code:: python

        convert_to_opus = processor.convert(mime_type='audio/opus')
        convert_to_opus(asset)

    :param function: Method to decorate
    :return: Configurable method
    """
    @functools.wraps(function)
    def wrapper(self, **kwargs: Any) -> Callable[..., Asset]:
        configured_operator = functools.partial(function, self, **kwargs)
        return configured_operator
    return wrapper


class OperatorError(Exception):
    """
    Represents an error that is raised whenever an error occurs in an
    :func:`~madam.core.operator`.
    """
    def __init__(self, *args):
        """
        Initializes a new `OperatorError`.
        """
        super().__init__(*args)


class Pipeline:
    """
    Represents a processing pipeline for :class:`~madam.core.Asset` objects.

    The pipeline can be configured to hold a list of asset processing
    operators, all of which are applied to one or more assets when calling the
    :func:`~madam.core.Pipeline.process` method.
    """
    def __init__(self) -> None:
        """
        Initializes a new pipeline without operators.
        """
        self.operators: MutableSequence[Callable] = []

    def process(self, *assets: Asset) -> Generator[Asset, float, None]:
        """
        Applies the operators in this pipeline on the specified assets.

        :param \\*assets: Asset objects to be processed
        :type \\*assets: Asset
        :return: Generator with processed assets
        """
        for asset in assets:
            processed_asset = asset
            for operator in self.operators:
                processed_asset = operator(processed_asset)
            yield processed_asset

    def add(self, operator: Callable) -> None:
        """
        Appends the specified operator to the processing chain.

        :param operator: Operator to be added
        """
        self.operators.append(operator)


class Processor(metaclass=abc.ABCMeta):
    """
    Represents an entity that can create :class:`~madam.core.Asset` objects
    from binary data.

    Every `Processor` needs to have an `__init__` method with an optional
    `config` parameter in order to be registered correctly.
    """
    @abc.abstractmethod
    def __init__(self, config: Optional[Mapping[str, Any]] = None) -> None:
        """
        Initializes a new `Processor`.

        :param config: Mapping with settings.
        """
        self.config: Dict[str, Any] = {}
        if config:
            self.config.update(config)

    @abc.abstractmethod
    def can_read(self, file: IO) -> bool:
        """
        Returns whether the specified MIME type is supported by this processor.

        :param file: file-like object to be tested
        :type file: IO
        :return: whether the data format of the specified file is supported or not
        :rtype: bool
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def read(self, file: IO) -> Asset:
        """
        Returns an :class:`~madam.core.Asset` object whose essence is identical to
        the contents of the specified file.

        :param file: file-like object to be read
        :type file: IO
        :return: Asset with essence
        :rtype: Asset
        :raises UnsupportedFormatError: if the specified data format is not supported
        """
        raise NotImplementedError()


class MetadataProcessor(metaclass=abc.ABCMeta):
    """
    Represents an entity that can manipulate metadata.

    Every `MetadataProcessor` needs to have an `__init__` method with an
    optional `config` parameter in order to be registered correctly.
    """
    @abc.abstractmethod
    def __init__(self, config: Optional[Mapping[str, Any]] = None) -> None:
        """
        Initializes a new `MetadataProcessor`.
        """
        self.config: Dict[str, Any] = {}
        if config:
            self.config.update(config)

    @property
    @abc.abstractmethod
    def formats(self) -> Iterable[str]:
        """
        The metadata formats which are supported.

        :return: supported metadata formats
        :rtype: set[str]
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def read(self, file: IO) -> Mapping:
        """
        Reads the file and returns the metadata.

        The metadata that is returned is grouped by type. The keys are specified by
        :attr:`~madam.core.MetadataProcessor.format`.

        :param file: File-like object to be read
        :type file: IO
        :return: Metadata contained in the file
        :rtype: Mapping
        :raises UnsupportedFormatError: if the data is corrupt or its format is not supported
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def strip(self, file: IO) -> IO:
        """
        Removes all metadata of the supported type from the specified file.

        :param file: file-like that should get stripped of the metadata
        :type file: IO
        :return: file-like object without metadata
        :rtype: IO
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def combine(self, file: IO, metadata: Mapping) -> IO:
        """
        Returns a byte stream whose contents represent the specified file where
        the specified metadata was added.

        :param metadata: Mapping of the metadata format to the metadata dict
        :type metadata: Mapping
        :param file: Container file
        :type file: IO
        :return: file-like object with combined content
        :rtype: IO
        """
        raise NotImplementedError()


class Madam:
    """
    Represents an instance of the library.
    """
    def __init__(self, config: Optional[Mapping[str, Any]] = None) -> None:
        """
        Initializes a new library instance with default configuration.

        The default configuration includes a list of all available Processor
        and MetadataProcessor implementations.

        :param config: Mapping with settings.
        """
        self.config: Dict[str, Any] = {}
        if config:
            self.config.update(config)

        # Initialize processors
        self.processors = {
            'madam.image.PillowProcessor',
            'madam.vector.SVGProcessor',
            'madam.ffmpeg.FFmpegProcessor',
        }
        self._processors = []
        for processor_path in set(self.processors):
            try:
                processor_class = Madam._import_from(processor_path)
            except ImportError:
                self.processors.remove(processor_path)
                continue
            processor = processor_class(self.config)
            self._processors.append(processor)

        # Initialize metadata processors
        self.metadata_processors = {
            'madam.exif.ExifMetadataProcessor',
            'madam.vector.SVGMetadataProcessor',
            'madam.ffmpeg.FFmpegMetadataProcessor',
        }
        self._metadata_processors = []
        for processor_path in set(self.metadata_processors):
            try:
                processor_class = Madam._import_from(processor_path)
            except ImportError:
                self.metadata_processors.remove(processor_path)
                continue
            processor = processor_class(self.config)
            self._metadata_processors.append(processor)

    @staticmethod
    def _import_from(member_path: str):
        """
        Returns the member located at the specified import path.

        :param member_path: Fully qualified name of the member to be imported
        :return: Member
        """
        module_path, member_name = member_path.rsplit('.', 1)
        module = importlib.import_module(module_path)
        member_class = getattr(module, member_name)
        return member_class

    def get_processor(self, file: IO) -> Optional[Processor]:
        """
        Returns a processor that can read the data in the specified file.
        If no suitable processor can be found None will be returned.

        :param file: file-like object to be parsed.
        :type file: IO
        :return: Processor object that can handle the data in the specified file,
                 or None if no suitable processor could be found.
        :rtype: Processor or None
        """
        for processor in self._processors:
            file.seek(0)
            if processor.can_read(file):
                file.seek(0)
                return processor
        return None

    def read(self, file: IO, additional_metadata: Mapping = None):
        r"""
        Reads the specified file and returns its contents as an :class:`~madam.core.Asset` object.

        :param file: file-like object to be parsed
        :type file: IO
        :param additional_metadata: optional metadata for the resulting asset.
               Existing metadata entries extracted from the file will be overwritten.
        :type additional_metadata: Mapping
        :returns: Asset representing the specified file
        :rtype: Asset
        :raises UnsupportedFormatError: if the file format cannot be recognized or is not supported
        :raises TypeError: if the file is None

        :Example:

        >>> import io
        >>> from madam import Madam
        >>> manager = Madam()
        >>> file = io.BytesIO(b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x01\x00\x00\x00\x01'
        ... b'\x08\x06\x00\x00\x00\x1f\x15\xc4\x89\x00\x00\x00\nIDATx\x9cc\x00\x01\x00\x00\x05\x00'
        ... b'\x01\r\n-\xb4\x00\x00\x00\x00IEND\xaeB`\x82')
        >>> asset = manager.read(file)
        """
        if not file:
            raise TypeError(f'Unable to read object of type {type(file)}')

        processor = self.get_processor(file)
        if not processor:
            raise UnsupportedFormatError()

        asset = processor.read(file)

        handled_formats = set()
        for metadata_processor in self._metadata_processors:
            asset_metadata = dict(asset.metadata)
            file.seek(0)
            try:
                metadata_by_format = metadata_processor.read(file)
                for metadata_format, metadata_values in metadata_by_format.items():
                    if metadata_format in handled_formats:
                        continue
                    asset_metadata[metadata_format] = metadata_values
                stripped_essence = metadata_processor.strip(asset.essence)
                clean_asset = Asset(stripped_essence, **asset_metadata)
                asset = clean_asset
                handled_formats.update(metadata_processor.formats)
            except UnsupportedFormatError:
                pass

        if additional_metadata:
            asset_metadata = dict(asset.metadata)
            asset_metadata.update(dict(additional_metadata))
            asset = Asset(asset.essence, **asset_metadata)

        return asset

    def write(self, asset: Asset, file: IO) -> None:
        r"""
        Write the :class:`~madam.core.Asset` object to the specified file.

        :param asset: Asset that contains the data to be written
        :type asset: Asset
        :param file: file-like object to be written
        :type file: IO

        :Example:

        >>> import io
        >>> import os
        >>> from madam import Madam
        >>> from madam.core import Asset
        >>> gif_asset = Asset(essence=io.BytesIO(b'GIF89a\x01\x00\x01\x00\x00\x00\x00;'), mime_type='image/gif')
        >>> manager = Madam()
        >>> with open(os.devnull, 'wb') as file:
        ...     manager.write(gif_asset, file)
        >>> wav_asset = Asset(
        ...     essence=io.BytesIO(b'RIFF$\x00\x00\x00WAVEfmt \x10\x00\x00\x00\x01\x00\x01\x00D\xac'
        ...             b'\x00\x00\x88X\x01\x00\x02\x00\x10\x00data\x00\x00\x00\x00'),
        ...     mime_type='video/mp4')
        >>> with open(os.devnull, 'wb') as file:
        ...     manager.write(wav_asset, file)
        """
        essence_with_metadata = asset.essence
        handled_formats = set()
        for metadata_processor in self._metadata_processors:
            metadata_by_format = {}

            for metadata_format in metadata_processor.formats:
                if metadata_format in handled_formats:
                    continue
                metadata = getattr(asset, metadata_format, None)
                if metadata is None:
                    handled_formats.add(metadata_format)
                    continue
                metadata_by_format[metadata_format] = metadata

            if not metadata_by_format:
                continue

            try:
                essence_with_metadata = metadata_processor.combine(essence_with_metadata, metadata_by_format)
                handled_formats.update(metadata_processor.formats)
            except UnsupportedFormatError:
                pass

        shutil.copyfileobj(essence_with_metadata, file)


AssetKey = TypeVar('AssetKey')
AssetTags = FrozenSet[str]


class AssetStorage(MutableMapping[AssetKey, Tuple[Asset, AssetTags]], Generic[AssetKey]):
    """
    Represents an abstract base class for data stores of
    :class:`~madam.core.Asset` objects.

    All implementations of `AssetStorage` require a constructor.

    The persistence guarantees for stored data may differ based on the
    respective storage implementation.
    """
    @abc.abstractmethod
    def __init__(self) -> None:
        """
        Initializes a new `AssetStorage`.
        """
        pass

    def filter(self, **kwargs: Any) -> Iterable[AssetKey]:
        """
        Returns a sequence of asset keys whose assets match the criteria that are
        specified by the passed arguments.

        :param \\**kwargs: Criteria defined as keys and values
        :return: Sequence of asset keys
        :rtype: Iterable
        """
        matches = []
        for asset_key, (asset, tags) in self.items():
            for key, value in kwargs.items():
                if asset.metadata.get(key) == value:
                    matches.append(asset_key)
        return matches

    def filter_by_tags(self, *tags: str) -> Iterable[AssetKey]:
        """
        Returns a set of all asset keys in this storage that have at least the
        specified tags.

        :param \\*tags: Mandatory tags of an asset to be included in result
        :return: Keys of the assets whose tags are a superset of the specified tags
        :rtype: Iterable
        """
        search_tags = frozenset(tags)
        return set(asset_key for asset_key, (asset, asset_tags) in self.items()
                   if search_tags <= asset_tags)


class InMemoryStorage(AssetStorage[Any]):
    """
    Represents a non-persistent storage backend for :class:`~madam.core.Asset`
    objects.

    Assets are not serialized, but stored in memory.
    """
    def __init__(self) -> None:
        """
        Initializes a new, empty `InMemoryStorage` object.
        """
        super().__init__()
        self.store: Dict[Any, Tuple[Asset, AssetTags]] = {}

    def __setitem__(self, asset_key: AssetKey, asset_and_tags: Tuple[Asset, AssetTags]):
        """
        Stores an :class:`~madam.core.Asset` in this asset storage using the
        specified key.

        The `asset_and_tags` argument is a tuple of the asset and the
        associated tags.

        Adding an asset key twice overwrites all tags for the asset.

        :param asset_key: Unique value used as a key to store the asset.
        :param asset_and_tags: Tuple of the asset and the tags associated with the asset
        :type asset_and_tags: Tuple[Asset, Set[str]]
        """
        asset, tags = asset_and_tags
        if not tags:
            tags = frozenset()
        self.store[asset_key] = asset, frozenset(tags)

    def __getitem__(self, asset_key: AssetKey) -> Tuple[Asset, AssetTags]:
        """
        Returns a tuple of the :class:`~madam.core.Asset` with the specified
        key and the tags associated with the asset.

        An error will be raised if the key does not exist.

        :param asset_key: Key of the asset for which the tags should be returned
        :return: A tuple containing an asset and a set of the tags associated with the asset
        :rtype: Tuple[Asset, Set[str]]
        :raise KeyError: if the key does not exist in this storage
        """
        if asset_key not in self.store:
            raise KeyError(f'Asset with key {asset_key!r} cannot be found in storage')
        return self.store[asset_key]

    def __delitem__(self, asset_key: AssetKey) -> None:
        """
        Removes the :class:`~madam.core.Asset` with the specified key from this
        asset storage, as well as all associated data (e.g. tags).

        :param asset_key: Key of the asset to be removed
        :raise KeyError: if the key does not exist in this storage
        """
        if asset_key not in self.store:
            raise KeyError(f'Asset with key {asset_key!r} cannot be found in storage')
        del self.store[asset_key]

    def __contains__(self, asset_key: AssetKey) -> bool:
        """
        Returns whether an asset with the specified key is stored in this
        asset storage.

        :param asset_key: Key of the asset that should be tested
        :return: `True` if the key exists, `False` otherwise
        :rtype: bool
        """
        return asset_key in self.store

    def __iter__(self) -> Iterator[AssetKey]:
        """
        Returns an object that can be used to iterate all asset that are stored
        in this asset storage.

        :return: Iterator object
        """
        return iter(list(self.store.keys()))

    def __len__(self) -> int:
        """
        Returns the number of assets in this storage.

        :return: Number of assets in this storage
        :rtype: int
        """
        return len(self.store)


class ShelveStorage(AssetStorage[str]):
    """
    Represents a persistent storage backend for :class:`~madam.core.Asset`
    objects. Asset keys must be strings.

    ShelveStorage uses a file on the file system to serialize Assets.
    """
    def __init__(self, path: Union[Path, str]):
        """
        Initializes a new `ShelveStorage` with the specified path.

        :param path: File system path where the data should be stored
        :type path: pathlib.Path or str
        """
        super().__init__()
        if os.path.exists(str(path)) and not os.path.isfile(str(path)):
            raise ValueError(f'The storage path {path!r} is not a file.')
        self.path = path

    def __setitem__(self, asset_key: str, asset_and_tags: Tuple[Asset, AssetTags]) -> None:
        """
        Stores an :class:`~madam.core.Asset` in this asset storage using the
        specified key.

        The `asset_and_tags` argument is a tuple of the asset and the
        associated tags.

        Adding an asset key twice overwrites all tags for the asset.

        :param asset_key: Unique value used as a key to store the asset.
        :param asset_and_tags: Tuple of the asset and the tags associated with the asset
        :type asset_and_tags: (Asset, collections.Iterable)
        """
        asset, tags = asset_and_tags
        if not tags:
            tags = frozenset()
        with shelve.open(str(self.path)) as store:
            store[asset_key] = asset, tags

    def __getitem__(self, asset_key: str) -> Tuple[Asset, AssetTags]:
        """
        Returns a tuple of the :class:`~madam.core.Asset` with the specified
        key and the tags associated with the asset.

        An error will be raised if the key does not exist.

        :param asset_key: Key of the asset for which the tags should be returned
        :type asset_key: str
        :return: A tuple containing an asset and a set of the tags associated with the asset
        :rtype: (Asset, set)
        :raise KeyError: if the key does not exist in this storage
        """
        with shelve.open(str(self.path)) as store:
            if asset_key not in store:
                raise KeyError(f'Asset with key {asset_key!r} cannot be found in storage')
            return store[asset_key]

    def __delitem__(self, asset_key: str) -> None:
        """
        Removes the :class:`~madam.core.Asset` with the specified key from this
        asset storage, as well as all associated data (e.g. tags).

        :param asset_key: Key of the asset to be removed
        :type asset_key: str
        :raise KeyError: if the key does not exist in this storage
        """
        with shelve.open(str(self.path)) as store:
            if asset_key not in store:
                raise KeyError(f'Asset with key {asset_key!r} cannot be found in storage')
            del store[asset_key]

    def __contains__(self, asset_key: object) -> bool:
        """
        Returns whether an asset with the specified key is stored in this
        asset storage.
        :param asset_key: Key of the asset that should be tested
        :type asset_key: str
        :return: `True` if the key exists, `False` otherwise
        :rtype: bool
        """
        if not isinstance(asset_key, str):
            return NotImplemented
        with shelve.open(str(self.path)) as store:
            return asset_key in store

    def __iter__(self) -> Iterator[str]:
        """
        Returns an object that can be used to iterate all asset that are stored
        in this asset storage.
        :return: Iterator object
        """
        with shelve.open(str(self.path)) as store:
            return iter(list(store.keys()))

    def __len__(self) -> int:
        """
        Returns the number of assets in this storage.
        :return: Number of assets in this storage
        :rtype: int
        """
        with shelve.open(str(self.path)) as store:
            return len(store)



import io
from enum import Enum
from typing import Any, Callable, IO, Mapping, Optional, Union

from bidict import bidict
import PIL.ExifTags
import PIL.Image

from madam.core import operator, OperatorError
from madam.core import Asset, Processor
from madam.mime import MimeType


class ResizeMode(Enum):
    """
    Represents a behavior for image resize operations.
    """
    #: Image exactly matches the specified dimensions
    EXACT = 0
    #: Image is resized to fit completely into the specified dimensions
    FIT = 1
    #: Image is resized to completely fill the specified dimensions
    FILL = 2


class FlipOrientation(Enum):
    """
    Represents an axis for image flip operations.
    """
    #: Horizontal axis
    HORIZONTAL = 0
    #: Vertical axis
    VERTICAL = 1


class PillowProcessor(Processor):
    """
    Represents a processor that uses Pillow as a backend.
    """
    __mime_type_to_pillow_type = bidict({
        MimeType('image/bmp'): 'BMP',
        MimeType('image/gif'): 'GIF',
        MimeType('image/jpeg'): 'JPEG',
        MimeType('image/png'): 'PNG',
        MimeType('image/tiff'): 'TIFF',
        MimeType('image/webp'): 'WEBP',
    })

    __format_defaults = {
        MimeType('image/gif'): dict(
            optimize=True,
        ),
        MimeType('image/jpeg'): dict(
            optimize=True,
            progressive=True,
            quality=80,
        ),
        MimeType('image/png'): dict(
            optimize=True,
        ),
        MimeType('image/tiff'): dict(
            compression='tiff_deflate',
        ),
        MimeType('image/webp'): dict(
            method=6,
            quality=80,
        ),
    }

    __pillow_mode_to_color_mode = bidict({
        '1': ('LUMA', 1, 'uint'),
        'L': ('LUMA', 8, 'uint'),
        'LA': ('LUMAA', 8, 'uint'),
        'P': ('PALETTE', 8, 'uint'),
        'RGB': ('RGB', 8, 'uint'),
        'RGBA': ('RGBA', 8, 'uint'),
        'RGBX': ('RGBX', 8, 'uint'),
        'CMYK': ('CMYK', 8, 'uint'),
        'YCbCr': ('YCbCr', 8, 'uint'),
        'LAB': ('LAB', 8, 'uint'),
        'HSV': ('HSV', 8, 'uint'),
        'I;16': ('LUMA', 16, 'uint'),
        'I': ('LUMA', 32, 'uint'),
        'F': ('LUMA', 32, 'float'),
    })

    def __init__(self, config: Optional[Mapping[str, Any]] = None) -> None:
        """
        Initializes a new `PillowProcessor`.

        :param config: Mapping with settings.
        """
        super().__init__(config)

    def read(self, file: IO) -> Asset:
        with PIL.Image.open(file) as image:
            mime_type = PillowProcessor.__mime_type_to_pillow_type.inv[image.format]
            color_space, bit_depth, data_type = PillowProcessor.__pillow_mode_to_color_mode[image.mode]
            metadata = dict(
                mime_type=str(mime_type),
                width=image.width,
                height=image.height,
                color_space=color_space,
                depth=bit_depth,
                data_type=data_type,
            )
        file.seek(0)
        asset = Asset(file, **metadata)
        return asset

    def can_read(self, file: IO) -> bool:
        try:
            PIL.Image.open(file)
            return True
        except IOError:
            return False
        finally:
            file.seek(0)

    @operator
    def resize(self, asset: Asset, width: int, height: int, mode: ResizeMode = ResizeMode.EXACT) -> Asset:
        """
        Creates a new Asset whose essence is resized according to the specified parameters.

        :param asset: Asset to be resized
        :type asset: Asset
        :param width: target width
        :type width: int
        :param height: target height
        :type height: int
        :param mode: resize behavior
        :type mode: ResizeMode
        :return: Asset with resized essence
        :rtype: Asset
        """
        mime_type = MimeType(asset.mime_type)
        with PIL.Image.open(asset.essence) as image:
            if mode == ResizeMode.EXACT:
                resized_width = width
                resized_height = height
            else:
                aspect = asset.width / asset.height
                aspect_target = width / height
                if mode == ResizeMode.FIT and aspect >= aspect_target or \
                   mode == ResizeMode.FILL and aspect <= aspect_target:
                    resize_factor = width / image.width
                else:
                    resize_factor = height / image.height
                resized_width = max(1, round(resize_factor * image.width))
                resized_height = max(1, round(resize_factor * image.height))
            # Pillow supports resampling only for 8-bit images
            resampling_method = PIL.Image.LANCZOS if asset.depth == 8 else PIL.Image.NEAREST
            resized_image = image.resize((resized_width, resized_height),
                                         resample=resampling_method)
        with resized_image:
            resized_asset = self._image_to_asset(resized_image, mime_type=mime_type)
        return resized_asset

    def _image_to_asset(self, image: PIL.Image.Image, mime_type: Union[MimeType, str]) -> Asset:
        """
        Converts an PIL image to a MADAM asset. The conversion can also include
        a change in file type.

        :param image: PIL image
        :type image: PIL.Image.Image
        :param mime_type: MIME type of the target asset
        :type mime_type: MimeType or str
        :return: MADAM asset with the specified MIME type
        :rtype: Asset
        """
        mime_type = MimeType(mime_type)

        pil_format = PillowProcessor.__mime_type_to_pillow_type[mime_type]
        pil_options = dict(PillowProcessor.__format_defaults.get(mime_type, {}))
        format_config = dict(self.config.get(mime_type.type, {}))
        format_config.update(self.config.get(str(mime_type), {}))

        image_buffer = io.BytesIO()

        if mime_type == MimeType('image/png') and image.mode != 'P':
            use_zopfli = format_config.get('zopfli', False)
            if use_zopfli:
                import zopfli
                zopfli_png = zopfli.ZopfliPNG()
                # Convert 16-bit per channel images to 8-bit per channel
                zopfli_png.lossy_8bit = False
                # Allow altering hidden colors of fully transparent pixels
                zopfli_png.lossy_transparent = True
                # Use all available optimization strategies
                zopfli_png.filter_strategies = format_config.get('zopfli_strategies', '0me')

                pil_options.pop('optimize', False)
                essence = io.BytesIO()
                image.save(essence, 'PNG', optimize=False, **pil_options)
                essence.seek(0)
                optimized_data = zopfli_png.optimize(essence.read())
                image_buffer.write(optimized_data)
            else:
                image.save(image_buffer, pil_format, **pil_options)
        elif mime_type == MimeType('image/jpeg'):
            pil_options['progressive'] = int(format_config.get('progressive', pil_options['progressive']))
            pil_options['quality'] = int(format_config.get('quality', pil_options['quality']))
            image.save(image_buffer, pil_format, **pil_options)
        elif mime_type == MimeType('image/tiff') and image.mode == 'P':
            pil_options.pop('compression', '')
            image.save(image_buffer, pil_format, **pil_options)
        elif mime_type == MimeType('image/webp'):
            pil_options['method'] = int(format_config.get('method', pil_options['method']))
            pil_options['quality'] = int(format_config.get('quality', pil_options['quality']))
            image.save(image_buffer, pil_format, **pil_options)
        else:
            image.save(image_buffer, pil_format, **pil_options)

        image_buffer.seek(0)

        asset = self.read(image_buffer)
        return asset

    def _rotate(self, asset: Asset, rotation: int) -> Asset:
        """
        Creates a new image asset from specified asset whose essence is rotated
        by the specified rotation.

        :param asset: Image asset to be rotated
        :type asset: Asset
        :param rotation: One of `PIL.Image.FLIP_LEFT_RIGHT`,
        `PIL.Image.FLIP_TOP_BOTTOM`, `PIL.Image.ROTATE_90`,
        `PIL.Image.ROTATE_180`, `PIL.Image.ROTATE_270`, or
        `PIL.Image.TRANSPOSE`
        :return: New image asset with rotated essence
        :rtype: Asset
        """
        mime_type = MimeType(asset.mime_type)
        with PIL.Image.open(asset.essence) as image:
            transposed_image = image.transpose(rotation)
        with transposed_image:
            transposed_asset = self._image_to_asset(transposed_image, mime_type=mime_type)
        return transposed_asset

    @operator
    def transpose(self, asset: Asset) -> Asset:
        """
        Creates a new image asset whose essence is the transpose of the
        specified asset's essence.

        :param asset: Image asset whose essence is to be transposed
        :type asset: Asset
        :return: New image asset with transposed essence
        :rtype: Asset
        """
        return self._rotate(asset, PIL.Image.TRANSPOSE)

    @operator
    def flip(self, asset: Asset, orientation: FlipOrientation) -> Asset:
        """
        Creates a new asset whose essence is flipped according the specified orientation.

        :param asset: Asset whose essence is to be flipped
        :type asset: Asset
        :param orientation: axis of the flip operation
        :type orientation: FlipOrientation
        :return: Asset with flipped essence
        :rtype: Asset
        """
        if orientation == FlipOrientation.HORIZONTAL:
            flip_orientation = PIL.Image.FLIP_LEFT_RIGHT
        else:
            flip_orientation = PIL.Image.FLIP_TOP_BOTTOM
        return self._rotate(asset, flip_orientation)

    @operator
    def auto_orient(self, asset: Asset) -> Asset:
        """
        Creates a new asset whose essence is rotated according to the Exif
        orientation. If no orientation metadata exists or asset is not rotated,
        an identical asset object is returned.

        :param asset: Asset with orientation metadata
        :type asset: Asset
        :return: Asset with rotated essence
        :rtype: Asset
        """
        orientation = asset.metadata.get('exif', {}).get('orientation')
        if orientation is None or orientation == 1:
            return asset

        flip_horizontally: Callable[[Asset], Asset] = self.flip(orientation=FlipOrientation.HORIZONTAL)
        flip_vertically: Callable[[Asset], Asset] = self.flip(orientation=FlipOrientation.VERTICAL)

        if orientation == 2:
            oriented_asset = flip_horizontally(asset)
        elif orientation == 3:
            oriented_asset = self._rotate(asset, PIL.Image.ROTATE_180)
        elif orientation == 4:
            oriented_asset = flip_vertically(asset)
        elif orientation == 5:
            oriented_asset = flip_vertically(self._rotate(asset, PIL.Image.ROTATE_90))
        elif orientation == 6:
            oriented_asset = self._rotate(asset, PIL.Image.ROTATE_270)
        elif orientation == 7:
            oriented_asset = flip_horizontally(self._rotate(asset, PIL.Image.ROTATE_90))
        elif orientation == 8:
            oriented_asset = self._rotate(asset, PIL.Image.ROTATE_90)
        else:
            raise OperatorError(f'Unable to correct image orientation with value {orientation}')

        return oriented_asset

    @operator
    def convert(self, asset: Asset, mime_type: Union[MimeType, str],
                color_space: Optional[str] = None, depth: Optional[int] = None,
                data_type: Optional[str] = None) -> Asset:
        """
        Creates a new asset of the specified MIME type from the essence of the
        specified asset.

        :param asset: Asset whose contents will be converted
        :type asset: Asset
        :param mime_type: Target MIME type
        :type mime_type: MimeType or str
        :param color_space: Name of color space
        :type color_space: str or None
        :param depth: Bit depth per channel
        :type depth: int or None
        :param data_type: Data type of the pixels, e.g. 'uint' or 'float'
        :type data_type: str or None
        :return: New asset with converted essence
        :rtype: Asset
        """
        mime_type = MimeType(mime_type)
        try:
            with PIL.Image.open(asset.essence) as image:
                color_mode = color_space or asset.color_space, depth or asset.depth, data_type or asset.data_type
                pil_mode = PillowProcessor.__pillow_mode_to_color_mode.inv.get(color_mode)
                if pil_mode is not None and pil_mode != image.mode:
                    image = image.convert(pil_mode)
                converted_asset = self._image_to_asset(image, mime_type)
        except (IOError, KeyError) as pil_error:
            raise OperatorError(f'Could not convert image to {mime_type}: {pil_error}')

        return converted_asset

    @operator
    def crop(self, asset: Asset, x: int, y: int, width: int, height: int) -> Asset:
        """
        Creates a new asset whose essence is cropped to the specified
        rectangular area.

        :param asset: Asset whose contents will be cropped
        :type asset: Asset
        :param x: horizontal offset of the cropping area from left
        :type x: int
        :param y: vertical offset of the cropping area from top
        :type y: int
        :param width: width of the cropping area
        :type width: int
        :param height: height of the cropping area
        :type height: int
        :return: New asset with cropped essence
        :rtype: Asset
        """
        if x == 0 and y == 0 and width == asset.width and height == asset.height:
            return asset

        max_x = max(0, min(asset.width, width + x))
        max_y = max(0, min(asset.height, height + y))
        min_x = max(0, min(asset.width, x))
        min_y = max(0, min(asset.height, y))

        if min_x == asset.width or min_y == asset.height or max_x <= min_x or max_y <= min_y:
            raise OperatorError(f'Invalid cropping area: <x={x!r}, y={y!r}, width={width!r}, height={height!r}>')

        with PIL.Image.open(asset.essence) as image:
            cropped_image = image.crop(box=(min_x, min_y, max_x, max_y))
        with cropped_image:
            cropped_asset = self._image_to_asset(cropped_image, mime_type=asset.mime_type)

        return cropped_asset

    @operator
    def rotate(self, asset: Asset, angle: float, expand: bool = False) -> Asset:
        """
        Creates an asset whose essence is rotated by the specified angle in
        degrees.

        :param asset: Asset whose contents will be rotated
        :type asset: Asset
        :param angle: Angle in degrees, counter clockwise
        :type angle: float
        :param expand: If true, changes the dimensions of the new asset so it
            can hold the entire rotated essence, otherwise the dimensions of
            the original asset will be used.
        :type expand: bool
        :return: New asset with rotated essence
        :rtype: Asset
        """
        if angle % 360.0 == 0.0:
            return asset

        with PIL.Image.open(asset.essence) as image:
            rotated_image = image.rotate(angle=angle, resample=PIL.Image.BICUBIC, expand=expand)
        with rotated_image:
            rotated_asset = self._image_to_asset(rotated_image, mime_type=asset.mime_type)

        return rotated_asset



from functools import total_ordering
from typing import Optional, Union


@total_ordering
class MimeType:
    """
    Represents a MIME type according to RFC 2045. This class behaves identical
    to its string representation in ``dict`` and ``set``.

    Limitations:

    - Suffixes are considered a part of the subtype
    - Parameters like ``charset`` are not supported and will be treated as part
      of the subtype
    """
    __slots__ = 'type', 'subtype'

    def __init__(self, mediatype: Optional[Union[str, 'MimeType']], subtype: Optional[str] = None):
        """
        Initializes a new MIME type with either

        - both, *media type* and *subtype* strings
        - a complete MIME type string like ``'audio/opus'``
        - another ``MimeType`` object

        :param mediatype: Can define either the media type, or the complete MIME
            type by passing a MIME type string or another ``MimeType`` instance.
        :type mediatype: str or MimeType or None
        :param subtype: Defines the subtype.
        :type subtype: str or None
        """
        self.type = None
        self.subtype = None
        if isinstance(mediatype, MimeType):
            if subtype is not None:
                raise ValueError('Cannot pass MimeType object and subtype string for initialization.')
            self.type = mediatype.type
            self.subtype = mediatype.subtype
        elif isinstance(mediatype, str):
            if mediatype:
                delimiter_count = mediatype.count('/')
                if delimiter_count:
                    if subtype is not None:
                        raise ValueError('Cannot pass MIME type string and subtype string for initialization.')
                    if delimiter_count > 1:
                        raise ValueError(f'Too many delimiters in {mediatype!r}')
                    mediatype, subtype = mediatype.split('/')
                if mediatype != '*':
                    self.type = str(mediatype).lower()
        elif mediatype is not None:
            raise TypeError(f'{type(mediatype).__qualname__!r} type is not allowed for initialization of MIME type')

        if isinstance(subtype, str):
            if subtype and subtype != '*':
                if '/' in subtype:
                    raise ValueError('Subtype cannot contain delimiters')
                self.subtype = str(subtype).lower()
        elif subtype is not None:
            raise TypeError(f'{type(subtype).__qualname__!r} type is not allowed for initialization of MIME subtype')

    def __str__(self) -> str:
        return '/'.join((self.type or '*', self.subtype or '*'))

    def __repr__(self) -> str:
        return f'{self.__class__.__qualname__}(mediatype={self.type!r}, subtype={self.subtype!r})'

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, str):
            return self == MimeType(other)
        if isinstance(other, MimeType):
            return self.type == other.type and self.subtype == other.subtype
        return NotImplemented

    def __lt__(self, other: object) -> bool:
        if isinstance(other, str):
            return self < MimeType(other)
        if isinstance(other, MimeType):
            if self.type == other.type:
                if self.subtype == other.subtype:
                    return False
                else:
                    if self.subtype is None:
                        return True
                    return str(self.subtype) < str(other.subtype)
            else:
                if self.type is None:
                    return True
                return str(self.type) < str(other.type)
        return NotImplemented



import datetime
import io
import shutil
import tempfile
from fractions import Fraction
from typing import Any, Callable, Dict, IO, Iterable, Mapping, Optional, Tuple

import piexif
from bidict import bidict

from madam.core import MetadataProcessor, UnsupportedFormatError
from madam.mime import MimeType


def _convert_sequence(dec_enc: Tuple[Callable, Callable]) -> Tuple[Callable, Callable]:
    return lambda exif_values: tuple(map(dec_enc[0], exif_values)), \
           lambda values: list(map(dec_enc[1], values))


def _convert_first(dec_enc: Tuple[Callable, Callable]) -> Tuple[Callable, Callable]:
    return lambda exif_values: dec_enc[0](exif_values[0]), \
           lambda value: [dec_enc[1](value)]


def _convert_mapping(mapping: Mapping) -> Tuple[Callable, Callable]:
    bidi = bidict(mapping)
    return lambda exif_value: bidi[exif_value], \
           lambda value: bidi.inv[value]


class ExifMetadataProcessor(MetadataProcessor):
    """
    Represents a metadata processor for Exif metadata.
    """
    supported_mime_types = {
        MimeType('image/jpeg'),
        MimeType('image/webp'),
    }

    metadata_to_exif = bidict({
        'aperture': ('Exif', piexif.ExifIFD.ApertureValue),
        'artist': ('0th', piexif.ImageIFD.Artist),
        'brightness': ('Exif', piexif.ExifIFD.BrightnessValue),
        'camera.manufacturer': ('0th', piexif.ImageIFD.Make),
        'camera.model': ('0th', piexif.ImageIFD.Model),
        'description': ('0th', piexif.ImageIFD.ImageDescription),
        'exposure_time': ('Exif', piexif.ExifIFD.ExposureTime),
        'firmware': ('0th', piexif.ImageIFD.Software),
        'fnumber': ('Exif', piexif.ExifIFD.FNumber),
        'focal_length': ('Exif', piexif.ExifIFD.FocalLength),
        'focal_length_35mm': ('Exif', piexif.ExifIFD.FocalLengthIn35mmFilm),
        'gps.altitude': ('GPS', piexif.GPSIFD.GPSAltitude),
        'gps.altitude_ref': ('GPS', piexif.GPSIFD.GPSAltitudeRef),
        'gps.latitude': ('GPS', piexif.GPSIFD.GPSLatitude),
        'gps.latitude_ref': ('GPS', piexif.GPSIFD.GPSLatitudeRef),
        'gps.longitude': ('GPS', piexif.GPSIFD.GPSLongitude),
        'gps.longitude_ref': ('GPS', piexif.GPSIFD.GPSLongitudeRef),
        'gps.map_datum': ('GPS', piexif.GPSIFD.GPSMapDatum),
        'gps.speed': ('GPS', piexif.GPSIFD.GPSSpeed),
        'gps.speed_ref': ('GPS', piexif.GPSIFD.GPSSpeedRef),
        'gps.date_stamp': ('GPS', piexif.GPSIFD.GPSDateStamp),
        'gps.time_stamp': ('GPS', piexif.GPSIFD.GPSTimeStamp),
        'lens.manufacturer': ('Exif', piexif.ExifIFD.LensMake),
        'lens.model': ('Exif', piexif.ExifIFD.LensModel),
        'orientation': ('0th', piexif.ImageIFD.Orientation),
        'shutter_speed': ('Exif', piexif.ExifIFD.ShutterSpeedValue),
        'software': ('0th', piexif.ImageIFD.ProcessingSoftware),
    })

    __STRING = lambda exif_val: exif_val.decode('utf-8'), lambda value: value.encode('utf-8')
    __INT = int, int
    __RATIONAL = lambda exif_val: float(Fraction(*exif_val)), \
                 lambda value: (Fraction(value).limit_denominator().numerator,
                                Fraction(value).limit_denominator().denominator)
    __DATE = lambda exif_val: datetime.datetime.strptime(exif_val.decode('utf-8'), '%Y:%m:%d').date(), \
             lambda value: value.strftime('%Y:%m:%d')
    __TIME = lambda exif_val: datetime.time(*map(lambda v: round(float(Fraction(*v))), exif_val)), \
             lambda value: ((value.hour, 1), (value.minute, 1), (value.second, 1))

    converters: Dict[str, Tuple[Callable, Callable]] = {
        'aperture': __RATIONAL,
        'artist': __STRING,
        'brightness': __RATIONAL,
        'camera.manufacturer': __STRING,
        'camera.model': __STRING,
        'description': __STRING,
        'exposure_time': __RATIONAL,
        'firmware': __STRING,
        'fnumber': __RATIONAL,
        'focal_length': __RATIONAL,
        'focal_length_35mm': __INT,
        'gps.altitude': __RATIONAL,
        'gps.altitude_ref': _convert_mapping({0: 'm_above_sea_level', 1: 'm_below_sea_level'}),
        'gps.latitude': _convert_sequence(__RATIONAL),
        'gps.latitude_ref': _convert_mapping({b'N': 'north', b'S': 'south'}),
        'gps.longitude': _convert_sequence(__RATIONAL),
        'gps.longitude_ref': _convert_mapping({b'E': 'east', b'W': 'west'}),
        'gps.map_datum': __STRING,
        'gps.speed': __RATIONAL,
        'gps.speed_ref': _convert_mapping({b'K': 'km/h', b'M': 'mph', b'N': 'kn'}),
        'gps.date_stamp': __DATE,
        'gps.time_stamp': __TIME,
        'lens.manufacturer': __STRING,
        'lens.model': __STRING,
        'orientation': __INT,
        'shutter_speed': __RATIONAL,
        'software': __STRING,
    }

    def __init__(self, config: Optional[Mapping[str, Any]] = None) -> None:
        """
        Initializes a new `ExifMetadataProcessor`.

        :param config: Mapping with settings
        """
        super().__init__(config)

    @property
    def formats(self) -> Iterable[str]:
        return {'exif'}

    def read(self, file: IO) -> Mapping[str, Mapping]:
        with tempfile.NamedTemporaryFile(mode='wb') as tmp:
            tmp.write(file.read())
            tmp.flush()

            try:
                metadata = piexif.load(tmp.name)
            except (piexif.InvalidImageDataError, ValueError):
                raise UnsupportedFormatError('Unsupported file format.')

        metadata_by_format = {}
        for metadata_format in self.formats:
            format_metadata = {}
            for ifd_key, ifd_values in metadata.items():
                if not isinstance(ifd_values, dict):
                    continue
                for exif_key, exif_value in ifd_values.items():
                    madam_key = ExifMetadataProcessor.metadata_to_exif.inv.get((ifd_key, exif_key))
                    if madam_key is None:
                        continue
                    convert_to_madam, _ = ExifMetadataProcessor.converters[madam_key]
                    format_metadata[madam_key] = convert_to_madam(exif_value)
            if format_metadata:
                metadata_by_format[metadata_format] = format_metadata
        return metadata_by_format

    def strip(self, file: IO) -> IO:
        result = io.BytesIO()
        with tempfile.NamedTemporaryFile(mode='w+b') as tmp:
            tmp.write(file.read())
            tmp.flush()

            try:
                metadata = piexif.load(tmp.name)
                if any(metadata.values()):
                    open(tmp.name, 'rb').read()
                    piexif.remove(tmp.name)
            except (piexif.InvalidImageDataError, ValueError, UnboundLocalError):
                raise UnsupportedFormatError('Unsupported file format.')

            tmp.seek(0)
            shutil.copyfileobj(tmp, result)
            result.seek(0)

        return result

    def combine(self, essence: IO, metadata_by_format: Mapping[str, Mapping]) -> IO:
        result = io.BytesIO()
        with tempfile.NamedTemporaryFile(mode='w+b') as tmp:
            tmp.write(essence.read())
            tmp.flush()

            try:
                exif_metadata = piexif.load(tmp.name)
            except (piexif.InvalidImageDataError, ValueError):
                raise UnsupportedFormatError('Unsupported essence format.')

            for metadata_format, metadata in metadata_by_format.items():
                if metadata_format not in self.formats:
                    raise UnsupportedFormatError(f'Metadata format {metadata_format!r} is not supported.')
                for madam_key, madam_value in metadata.items():
                    if madam_key not in ExifMetadataProcessor.metadata_to_exif:
                        continue
                    ifd_key, exif_key = ExifMetadataProcessor.metadata_to_exif[madam_key]
                    if ifd_key not in exif_metadata:
                        exif_metadata[ifd_key] = {}
                    _, convert_to_exif = ExifMetadataProcessor.converters[madam_key]
                    exif_metadata[ifd_key][exif_key] = convert_to_exif(madam_value)

            try:
                piexif.insert(piexif.dump(exif_metadata), tmp.name)
            except (piexif.InvalidImageDataError, ValueError):
                raise UnsupportedFormatError(f'Could not write metadata: {metadata_by_format!r}')

            tmp.seek(0)
            shutil.copyfileobj(tmp, result)
            result.seek(0)

        return result



import io
import json
import multiprocessing
import os
import shutil
import subprocess
import tempfile
from collections import namedtuple
from math import ceil, cos, pi, radians, sin
from typing import Any, Dict, IO, Iterable, Mapping, MutableMapping, Optional, Sequence, Tuple, Union

from bidict import bidict

from madam.core import Asset, MetadataProcessor, Processor, operator, OperatorError, UnsupportedFormatError
from madam.mime import MimeType


def _probe(file: IO) -> Any:
    with tempfile.NamedTemporaryFile(mode='wb') as temp_in:
        shutil.copyfileobj(file, temp_in.file)  # type: ignore
        temp_in.flush()
        file.seek(0)

        command = 'ffprobe -loglevel error -print_format json -show_format -show_streams'.split()
        command.append(temp_in.name)
        result = subprocess.run(command, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE, check=True)

    string_result = result.stdout.decode('utf-8')
    json_obj = json.loads(string_result)

    return json_obj


def _combine_metadata(asset, *cloned_keys: str, **additional_metadata: Any) -> MutableMapping[str, Any]:
    metadata = {key: asset.metadata[key] for key in cloned_keys if key in asset.metadata}
    metadata.update(additional_metadata)
    return metadata


_FFmpegMode = namedtuple('_FFmpegMode', 'name, component_count, bits_per_pixel, readable, writeable, '
                                        'hw_accelerated, paletted, bitstream')


def _get_decoder_and_stream_type(probe_data: Mapping[str, Any]) -> Tuple[str, str]:
    decoder_name = probe_data['format']['format_name']

    stream_types = {stream['codec_type'] for stream in probe_data['streams']}
    if 'video' in stream_types:
        stream_type = 'video'
    elif 'audio' in stream_types:
        stream_type = 'audio'
    elif 'subtitle' in stream_types:
        stream_type = 'subtitle'
    else:
        stream_type = ''

    return decoder_name, stream_type


def _param_map_to_seq(param_mapping: Mapping[str, Any]) -> Sequence[str]:
    params = []
    for param, value in param_mapping.items():
        params.append('-{}'.format(param))
        if value is not None:
            params.append(str(value))
    return params


class _FFmpegContext(tempfile.TemporaryDirectory):
    def __init__(self, source: IO, result: IO) -> None:
        super().__init__(prefix='madam')
        self.__source = source
        self.__result = result

    def __enter__(self) -> '_FFmpegContext':
        tmpdir_path = super().__enter__()
        self.input_path = os.path.join(tmpdir_path, 'input_file')
        self.output_path = os.path.join(tmpdir_path, 'output_file')

        with open(self.input_path, 'wb') as temp_in:
            shutil.copyfileobj(self.__source, temp_in)
            self.__source.seek(0)

        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        if os.path.exists(self.output_path):
            with open(self.output_path, 'rb') as temp_out:
                shutil.copyfileobj(temp_out, self.__result)
                self.__result.seek(0)

        super().__exit__(exc_type, exc_val, exc_tb)


class FFmpegProcessor(Processor):
    """
    Represents a processor that uses FFmpeg to read audio and video data.

    The minimum version of FFmpeg required is v3.3.
    """

    __decoder_and_stream_type_to_mime_type = {
        ('matroska,webm', 'video'): MimeType('video/x-matroska'),
        ('mov,mp4,m4a,3gp,3g2,mj2', 'video'): MimeType('video/quicktime'),
        ('avi', 'video'): MimeType('video/x-msvideo'),
        ('mpegts', 'video'): MimeType('video/mp2t'),
        ('nut', 'video'): MimeType('video/x-nut'),
        ('ogg', 'video'): MimeType('video/ogg'),
        ('mp3', 'audio'): MimeType('audio/mpeg'),
        ('nut', 'audio'): MimeType('audio/x-nut'),
        ('ogg', 'audio'): MimeType('audio/ogg'),
        ('wav', 'audio'): MimeType('audio/wav'),
        ('webvtt', 'subtitle'): MimeType('text/vtt'),
    }

    __mime_type_to_encoder = {
        MimeType('video/x-matroska'): 'matroska',
        MimeType('video/quicktime'): 'mov',
        MimeType('video/x-msvideo'): 'avi',
        MimeType('video/mp2t'): 'mpegts',
        MimeType('video/x-nut'): 'nut',
        MimeType('video/ogg'): 'ogg',
        MimeType('audio/mpeg'): 'mp3',
        MimeType('audio/x-nut'): 'nut',
        MimeType('audio/ogg'): 'ogg',
        MimeType('audio/wav'): 'wav',
        MimeType('image/bmp'): 'image2',
        MimeType('image/gif'): 'gif',
        MimeType('image/jpeg'): 'image2',
        MimeType('image/png'): 'image2',
        MimeType('image/tiff'): 'image2',
        MimeType('image/webp'): 'image2',
        MimeType('text/vnd.dvb.subtitle'): 'dvbsub',
        MimeType('text/vtt'): 'webvtt',
    }

    __mime_type_to_codec = {
        MimeType('image/bmp'): 'bmp',
        MimeType('image/gif'): 'gif',
        MimeType('image/jpeg'): 'mjpeg',
        MimeType('image/png'): 'png',
        MimeType('image/tiff'): 'tiff',
        MimeType('image/webp'): 'libwebp',
    }

    __codec_options = {
        'video': {
            'libx264': {
                'preset': 'slow',
                'crf': 23,
            },
            'libx265': {
                'preset': 'slow',
                'crf': 28,
            },
            'libvpx': {
                'crf': 10,
            },
            'libvpx-vp9': {
                'row-mt': 1,
                'crf': 32,
            },
            'opus': {
                'strict': -2
            },
            'vorbis': {
                'ac': 2,
                'strict': -2
            },
            'vp9': {
                'tile-columns': 6,
                'crf': 32,
            },
        }
    }

    __container_options = {
        MimeType('video/x-matroska'): [
            '-avoid_negative_ts', 'make_zero',
        ],
    }

    __ffmpeg_pix_fmt_to_color_mode = {
        # Luminance
        'gray': ('LUMA', 8, 'uint'),
        'gray9be': ('LUMA', 9, 'uint'),
        'gray9le': ('LUMA', 9, 'uint'),
        'gray10be': ('LUMA', 10, 'uint'),
        'gray10le': ('LUMA', 10, 'uint'),
        'gray12be': ('LUMA', 12, 'uint'),
        'gray12le': ('LUMA', 12, 'uint'),
        'gray16be': ('LUMA', 16, 'uint'),
        'gray16le': ('LUMA', 16, 'uint'),
        # Luminance, alpha
        'monob': ('LUMA', 1, 'uint'),
        'monow': ('LUMA', 1, 'uint'),
        'ya8': ('LUMAA', 8, 'uint'),
        'ya16be': ('LUMAA', 16, 'uint'),
        'ya16le': ('LUMAA', 16, 'uint'),
        # Palette
        'pal8': ('PALETTE', 8, 'uint'),
        # Red, green, blue
        'bayer_bggr8': ('RGB', 8, 'uint'),
        'bayer_bggr16be': ('RGB', 16, 'uint'),
        'bayer_bggr16le': ('RGB', 16, 'uint'),
        'bayer_gbrg8': ('RGB', 8, 'uint'),
        'bayer_gbrg16be': ('RGB', 16, 'uint'),
        'bayer_gbrg16le': ('RGB', 16, 'uint'),
        'bayer_grbg8': ('RGB', 8, 'uint'),
        'bayer_grbg16be': ('RGB', 16, 'uint'),
        'bayer_grbg16le': ('RGB', 16, 'uint'),
        'bayer_rggb8': ('RGB', 8, 'uint'),
        'bayer_rggb16be': ('RGB', 16, 'uint'),
        'bayer_rggb16le': ('RGB', 16, 'uint'),
        'bgr4': ('RGB', 8, 'uint'),
        'bgr4_byte': ('RGB', 8, 'uint'),
        'bgr24': ('RGB', 8, 'uint'),
        'bgr48be': ('RGB', 16, 'uint'),
        'bgr48le': ('RGB', 16, 'uint'),
        'bgr444be': ('RGB', 8, 'uint'),
        'bgr444le': ('RGB', 8, 'uint'),
        'bgr555be': ('RGB', 8, 'uint'),
        'bgr555le': ('RGB', 8, 'uint'),
        'bgr565be': ('RGB', 8, 'uint'),
        'bgr565le': ('RGB', 8, 'uint'),
        'bgr8': ('RGB', 8, 'uint'),
        'gbrp': ('RGB', 8, 'uint'),
        'gbrp9be': ('RGB', 9, 'uint'),
        'gbrp9le': ('RGB', 9, 'uint'),
        'gbrp10be': ('RGB', 10, 'uint'),
        'gbrp10le': ('RGB', 10, 'uint'),
        'gbrp12be': ('RGB', 12, 'uint'),
        'gbrp12le': ('RGB', 12, 'uint'),
        'gbrp14be': ('RGB', 14, 'uint'),
        'gbrp14le': ('RGB', 14, 'uint'),
        'gbrp16be': ('RGB', 16, 'uint'),
        'gbrp16le': ('RGB', 16, 'uint'),
        'gbrpf32be': ('RGB', 32, 'float'),
        'gbrpf32le': ('RGB', 32, 'float'),
        'rgb4': ('RGB', 8, 'uint'),
        'rgb4_byte': ('RGB', 8, 'uint'),
        'rgb24': ('RGB', 8, 'uint'),
        'rgb48be': ('RGB', 16, 'uint'),
        'rgb48le': ('RGB', 16, 'uint'),
        'rgb444be': ('RGB', 8, 'uint'),
        'rgb444le': ('RGB', 8, 'uint'),
        'rgb555be': ('RGB', 8, 'uint'),
        'rgb555le': ('RGB', 8, 'uint'),
        'rgb565be': ('RGB', 8, 'uint'),
        'rgb565le': ('RGB', 8, 'uint'),
        'rgb8': ('RGB', 8, 'uint'),
        # Red, green, blue, alpha
        'abgr': ('RGBA', 8, 'uint'),
        'argb': ('RGBA', 8, 'uint'),
        'bgra': ('RGBA', 8, 'uint'),
        'bgra64be': ('RGBA', 16, 'uint'),
        'bgra64le': ('RGBA', 16, 'uint'),
        'gbrap': ('RGBA', 8, 'uint'),
        'gbrap10be': ('RGBA', 16, 'uint'),
        'gbrap10le': ('RGBA', 16, 'uint'),
        'gbrap12be': ('RGBA', 16, 'uint'),
        'gbrap12le': ('RGBA', 16, 'uint'),
        'gbrap16be': ('RGBA', 16, 'uint'),
        'gbrap16le': ('RGBA', 16, 'uint'),
        'gbrapf32be': ('RGBA', 32, 'float'),
        'gbrapf32le': ('RGBA', 32, 'float'),
        'rgba': ('RGBA', 8, 'uint'),
        'rgba64be': ('RGBA', 16, 'uint'),
        'rgba64le': ('RGBA', 16, 'uint'),
        # Red, green, blue, padding
        '0bgr': ('RGBX', 8, 'uint'),
        '0rgb': ('RGBX', 8, 'uint'),
        'bgr0': ('RGBX', 8, 'uint'),
        'rgb0': ('RGBX', 8, 'uint'),
        # X, Y, Z
        'xyz12be': ('XYZ', 12, 'uint'),
        'xyz12le': ('XYZ', 12, 'uint'),
        # Luminance, blue-difference chrominance, red-difference chrominance
        'nv12': ('YUV', 8, 'uint'),
        'nv16': ('YUV', 8, 'uint'),
        'nv20be': ('YUV', 8, 'uint'),
        'nv20le': ('YUV', 8, 'uint'),
        'nv21': ('YUV', 8, 'uint'),
        'p010be': ('YUV', 10, 'uint'),
        'p010le': ('YUV', 10, 'uint'),
        'p016be': ('YUV', 16, 'uint'),
        'p016le': ('YUV', 16, 'uint'),
        'uyvy422': ('YUV', 8, 'uint'),
        'uyyvyy411': ('YUV', 8, 'uint'),
        'yuv410p': ('YUV', 8, 'uint'),
        'yuv411p': ('YUV', 8, 'uint'),
        'yuv420p': ('YUV', 8, 'uint'),
        'yuv420p9be': ('YUV', 9, 'uint'),
        'yuv420p9le': ('YUV', 9, 'uint'),
        'yuv420p10be': ('YUV', 10, 'uint'),
        'yuv420p10le': ('YUV', 10, 'uint'),
        'yuv420p12be': ('YUV', 12, 'uint'),
        'yuv420p12le': ('YUV', 12, 'uint'),
        'yuv420p14be': ('YUV', 14, 'uint'),
        'yuv420p14le': ('YUV', 14, 'uint'),
        'yuv420p16be': ('YUV', 16, 'uint'),
        'yuv420p16le': ('YUV', 16, 'uint'),
        'yuv422p': ('YUV', 8, 'uint'),
        'yuv422p9be': ('YUV', 9, 'uint'),
        'yuv422p9le': ('YUV', 9, 'uint'),
        'yuv422p10be': ('YUV', 10, 'uint'),
        'yuv422p10le': ('YUV', 10, 'uint'),
        'yuv422p12be': ('YUV', 12, 'uint'),
        'yuv422p12le': ('YUV', 12, 'uint'),
        'yuv422p14be': ('YUV', 14, 'uint'),
        'yuv422p14le': ('YUV', 14, 'uint'),
        'yuv422p16be': ('YUV', 16, 'uint'),
        'yuv422p16le': ('YUV', 16, 'uint'),
        'yuv440p': ('YUV', 8, 'uint'),
        'yuv440p10be': ('YUV', 10, 'uint'),
        'yuv440p10le': ('YUV', 10, 'uint'),
        'yuv440p12be': ('YUV', 12, 'uint'),
        'yuv440p12le': ('YUV', 12, 'uint'),
        'yuv444p': ('YUV', 8, 'uint'),
        'yuv444p9be': ('YUV', 9, 'uint'),
        'yuv444p9le': ('YUV', 9, 'uint'),
        'yuv444p10be': ('YUV', 10, 'uint'),
        'yuv444p10le': ('YUV', 10, 'uint'),
        'yuv444p12be': ('YUV', 12, 'uint'),
        'yuv444p12le': ('YUV', 12, 'uint'),
        'yuv444p14be': ('YUV', 14, 'uint'),
        'yuv444p14le': ('YUV', 14, 'uint'),
        'yuv444p16be': ('YUV', 16, 'uint'),
        'yuv444p16le': ('YUV', 16, 'uint'),
        'yuvj411p': ('YUV', 8, 'uint'),
        'yuvj420p': ('YUV', 8, 'uint'),
        'yuvj422p': ('YUV', 8, 'uint'),
        'yuvj440p': ('YUV', 8, 'uint'),
        'yuvj444p': ('YUV', 8, 'uint'),
        'yuyv422': ('YUV', 8, 'uint'),
        'yvyu422': ('YUV', 8, 'uint'),
        # Luminance, blue-difference chrominance, red-difference chrominance, alpha
        'ayuv64be': ('YUVA', 16, 'uint'),
        'ayuv64le': ('YUVA', 16, 'uint'),
        'yuva420p': ('YUVA', 8, 'uint'),
        'yuva420p9be': ('YUVA', 9, 'uint'),
        'yuva420p9le': ('YUVA', 9, 'uint'),
        'yuva420p10be': ('YUVA', 10, 'uint'),
        'yuva420p10le': ('YUVA', 10, 'uint'),
        'yuva420p16be': ('YUVA', 16, 'uint'),
        'yuva420p16le': ('YUVA', 16, 'uint'),
        'yuva422p': ('YUVA', 8, 'uint'),
        'yuva422p9be': ('YUVA', 9, 'uint'),
        'yuva422p9le': ('YUVA', 9, 'uint'),
        'yuva422p10be': ('YUVA', 10, 'uint'),
        'yuva422p10le': ('YUVA', 10, 'uint'),
        'yuva422p16be': ('YUVA', 16, 'uint'),
        'yuva422p16le': ('YUVA', 16, 'uint'),
        'yuva444p': ('YUVA', 8, 'uint'),
        'yuva444p9be': ('YUVA', 9, 'uint'),
        'yuva444p9le': ('YUVA', 9, 'uint'),
        'yuva444p10be': ('YUVA', 10, 'uint'),
        'yuva444p10le': ('YUVA', 10, 'uint'),
        'yuva444p16be': ('YUVA', 16, 'uint'),
        'yuva444p16le': ('YUVA', 16, 'uint'),
    }

    __color_mode_to_ffmpeg_pix_fmt = {
        ('LUMA', 1, 'uint'): 'monob',
        ('LUMA', 8, 'uint'): 'gray',
        ('LUMA', 9, 'uint'): 'gray9le',
        ('LUMA', 10, 'uint'): 'gray10le',
        ('LUMA', 12, 'uint'): 'gray12le',
        ('LUMA', 16, 'uint'): 'gray16le',
        ('LUMAA', 8, 'uint'): 'ya8',
        ('LUMAA', 16, 'uint'): 'ya16le',
        ('PALETTE', 8, 'uint'): 'pal8',
        ('RGB', 8, 'uint'): 'rgb24',
        ('RGB', 9, 'uint'): 'gbrp9le',
        ('RGB', 10, 'uint'): 'gbrp10le',
        ('RGB', 12, 'uint'): 'gbrp12le',
        ('RGB', 16, 'uint'): 'rgb48le',
        ('RGB', 32, 'float'): 'gbrpf32le',
        ('RGBA', 8, 'uint'): 'rgba',
        ('RGBA', 16, 'uint'): 'rgba64le',
        ('RGBA', 32, 'float'): 'gbrapf32le',
        ('RGBX', 8, 'uint'): 'rgb0',
        ('XYZ', 12, 'uint'): 'xyz12le',
        ('YUV', 8, 'uint'): 'yuv420p',
        ('YUV', 9, 'uint'): 'yuv420p9le',
        ('YUV', 10, 'uint'): 'yuv420p10le',
        ('YUV', 12, 'uint'): 'yuv420p12le',
        ('YUV', 14, 'uint'): 'yuv420p14le',
        ('YUV', 16, 'uint'): 'yuv420p16le',
        ('YUVA', 8, 'uint'): 'yuva420p',
        ('YUVA', 9, 'uint'): 'yuva420p9le',
        ('YUVA', 10, 'uint'): 'yuva420p10le',
        ('YUVA', 16, 'uint'): 'yuva420p16le',
    }

    def __init__(self, config: Optional[Mapping[str, Any]] = None) -> None:
        """
        Initializes a new `FFmpegProcessor`.

        :param config: Mapping with settings.
        :raises EnvironmentError: if the installed version of ffprobe does not match the minimum version requirement
        """
        super().__init__(config)

        self._min_version = '3.3'
        command = 'ffprobe -version'.split()
        result = subprocess.run(command, stdout=subprocess.PIPE)
        string_result = result.stdout.decode('utf-8')
        version_string = string_result.split()[2]
        if version_string < self._min_version:
            raise EnvironmentError(f'Found ffprobe version {version_string}. '
                                   f'Requiring at least version {self._min_version}.')

        self.__threads = multiprocessing.cpu_count()

    def can_read(self, file: IO) -> bool:
        try:
            probe_data = _probe(file)
            decoder_and_stream_type = _get_decoder_and_stream_type(probe_data)
            mime_type = self.__decoder_and_stream_type_to_mime_type.get(decoder_and_stream_type)
            return bool(mime_type)
        except subprocess.CalledProcessError:
            return False

    def read(self, file: IO) -> Asset:
        try:
            probe_data = _probe(file)
        except subprocess.CalledProcessError:
            raise UnsupportedFormatError('Unsupported file format.')

        decoder_and_stream_type = _get_decoder_and_stream_type(probe_data)
        mime_type = self.__decoder_and_stream_type_to_mime_type.get(decoder_and_stream_type)
        if not mime_type:
            raise UnsupportedFormatError('Unsupported metadata source.')

        metadata: Dict[str, Any] = dict(
            mime_type=str(mime_type),
        )

        if 'duration' in probe_data['format']:
            metadata['duration'] = float(probe_data['format']['duration'])

        for stream in probe_data['streams']:
            stream_type = stream.get('codec_type')
            if stream_type in {'video', 'audio', 'subtitle'}:
                # Only use first stream
                if stream_type in metadata:
                    break
                metadata[stream_type] = {}
            if 'width' in stream:
                metadata['width'] = max(stream['width'], metadata.get('width', 0))
            if 'height' in stream:
                metadata['height'] = max(stream['height'], metadata.get('height', 0))
            if stream_type not in metadata:
                continue
            if 'codec_name' in stream:
                metadata[stream_type]['codec'] = stream['codec_name']
            if 'bit_rate' in stream:
                metadata[stream_type]['bitrate'] = float(stream['bit_rate'])/1000.0
            if 'pix_fmt' in stream:
                color_space, depth, data_type = FFmpegProcessor.__ffmpeg_pix_fmt_to_color_mode[stream['pix_fmt']]
                metadata[stream_type]['color_space'] = color_space
                metadata[stream_type]['depth'] = depth
                metadata[stream_type]['data_type'] = data_type

        return Asset(essence=file, **metadata)

    @operator
    def resize(self, asset: Asset, width: int, height: int) -> Asset:
        """
        Creates a new image or video asset of the specified width and height
        from the essence of the specified image or video asset.

        Width and height must be positive numbers.

        :param asset: Video asset that will serve as the source for the frame
        :type asset: Asset
        :param width: Width of the resized asset
        :type width: int
        :param height: Height of the resized asset
        :type height: int
        :return: New asset with specified width and height
        :rtype: Asset
        """
        if width < 1 or height < 1:
            raise ValueError(f'Invalid dimensions: {width:d}x{height:d}')

        mime_type = MimeType(asset.mime_type)
        encoder_name = self.__mime_type_to_encoder.get(mime_type)
        if not encoder_name:
            raise UnsupportedFormatError(f'Unsupported asset type: {mime_type}')
        if mime_type.type not in ('image', 'video'):
            raise OperatorError(f'Cannot resize asset of type {mime_type}')

        result = io.BytesIO()
        with _FFmpegContext(asset.essence, result) as ctx:
            with open(ctx.input_path, 'wb') as temp_in:
                shutil.copyfileobj(asset.essence, temp_in)
                temp_in.flush()

            command = ['ffmpeg', '-loglevel', 'error',
                       '-f', encoder_name, '-i', ctx.input_path,
                       '-filter:v', f'scale={width:d}:{height:d}',
                       '-threads', str(self.__threads),
                       '-f', encoder_name, '-y', ctx.output_path]

            try:
                subprocess.run(command, stderr=subprocess.PIPE, check=True)
            except subprocess.CalledProcessError as ffmpeg_error:
                error_message = ffmpeg_error.stderr.decode('utf-8')
                raise OperatorError(f'Could not resize asset: {error_message}')

        metadata = _combine_metadata(asset,
                                     'mime_type', 'duration', 'video', 'audio', 'subtitle',
                                     width=width, height=height)

        return Asset(essence=result, **metadata)

    @operator
    def convert(self, asset: Asset, mime_type: Union[MimeType, str],
                video: Optional[Mapping[str, Any]] = None,
                audio: Optional[Mapping[str, Any]] = None,
                subtitle: Optional[Mapping[str, Any]] = None) -> Asset:
        """
        Creates a new asset of the specified MIME type from the essence of the
        specified asset.

        Additional options can be specified for video, audio, and subtitle streams.
        Options are passed as dictionary instances and can contain various keys for
        each stream type.

        **Options for video streams:**

        - **codec** – Processor-specific name of the video codec as string
        - **bitrate** – Target bitrate in kBit/s as float number

        **Options for audio streams:**

        - **codec** – Processor-specific name of the audio codec as string
        - **bitrate** – Target bitrate in kBit/s as float number

        **Options for subtitle streams:**

        - **codec** – Processor-specific name of the subtitle format as string

        :param asset: Asset whose contents will be converted
        :type asset: Asset
        :param mime_type: MIME type of the video container
        :type mime_type: MimeType or str
        :param video: Dictionary with options for video streams.
        :type video: dict or None
        :param audio: Dictionary with options for audio streams.
        :type audio: dict or None
        :param subtitle: Dictionary with the options for subtitle streams.
        :type subtitle: dict or None
        :return: New asset with converted essence
        :rtype: Asset
        """
        mime_type = MimeType(mime_type)
        encoder_name = self.__mime_type_to_encoder.get(mime_type)
        if not encoder_name:
            raise UnsupportedFormatError(f'Unsupported asset type: {mime_type}')

        result = io.BytesIO()
        with _FFmpegContext(asset.essence, result) as ctx:
            command = [
                'ffmpeg',
                '-loglevel', 'error',
                '-i', ctx.input_path
            ]

            format_config = dict(self.config.get(mime_type.type, {}))
            if mime_type.type == 'video':
                keyframe_interval = int(format_config.get('keyframe_interval', 100))
                command.extend(['-g', str(keyframe_interval)])

            if video:
                if 'codec' in video:
                    if video['codec']:
                        command.extend(['-c:v', video['codec']])
                        codec_options = dict(FFmpegProcessor.__codec_options.get('video', {}).get(video['codec'], []))
                        codec_config = self.config.get(f'codec/{video["codec"]}', {})
                        if 'crf' in codec_config:
                            codec_options['crf'] = int(codec_config['crf'])
                        command.extend(_param_map_to_seq(codec_options))
                    else:
                        command.extend(['-vn'])
                if video.get('bitrate'):
                    command.extend([
                        '-b:v', f'{video["bitrate"]:d}k'
                    ])
                if video.get('color_space') or video.get('depth') or video.get('data_type'):
                    color_mode = (
                        video.get('color_space', asset.video.get('color_space', 'YUV')),
                        video.get('depth', asset.video.get('depth', 8)),
                        video.get('data_type', asset.video.get('data_type', 'uint')),
                    )
                    ffmpeg_pix_fmt = FFmpegProcessor.__color_mode_to_ffmpeg_pix_fmt.get(color_mode)
                    if ffmpeg_pix_fmt:
                        command.extend(['-pix_fmt', ffmpeg_pix_fmt])
            if audio:
                if 'codec' in audio:
                    if audio['codec']:
                        command.extend(['-c:a', audio['codec']])
                        codec_options = FFmpegProcessor.__codec_options.get('audio', {}).get(audio['codec'], [])
                        command.extend(codec_options)
                    else:
                        command.extend(['-an'])
                if audio.get('bitrate'):
                    command.extend(['-b:a', f'{audio["bitrate"]:d}k'])
            if subtitle:
                if 'codec' in subtitle:
                    if subtitle['codec']:
                        command.extend(['-c:s', subtitle['codec']])
                        codec_options = FFmpegProcessor.__codec_options.get('subtitles', {})
                        command.extend(codec_options.get(subtitle['codec'], []))
                    else:
                        command.extend(['-sn'])

            container_options = FFmpegProcessor.__container_options.get(mime_type, [])
            container_config = self.config.get(str(mime_type), {})
            if mime_type == 'video/quicktime':
                use_faststart = container_config.get('faststart', True)
                if use_faststart:
                    container_options.extend(['-movflags', '+faststart'])
            command.extend(container_options)

            command.extend(['-threads', str(self.__threads),
                            '-f', encoder_name, '-y', ctx.output_path])

            try:
                subprocess.run(command, stderr=subprocess.PIPE, check=True)
            except subprocess.CalledProcessError as ffmpeg_error:
                error_message = ffmpeg_error.stderr.decode('utf-8')
                raise OperatorError(f'Could not convert asset: {error_message}')

        return self.read(result)

    @operator
    def trim(self, asset: Asset, from_seconds: float = 0, to_seconds: float = 0) -> Asset:
        """
        Creates a trimmed audio or video asset that only contains the data
        between from_seconds and to_seconds.

        :param asset: Audio or video asset, which will serve as the source
        :type asset: Asset
        :param from_seconds: Start time of the clip in seconds
        :type from_seconds: float
        :param to_seconds: End time of the clip in seconds
        :type to_seconds: float
        :return: New asset with trimmed essence
        :rtype: Asset
        """
        mime_type = MimeType(asset.mime_type)
        encoder_name = self.__mime_type_to_encoder.get(mime_type)
        if not encoder_name or mime_type.type not in ('audio', 'video'):
            raise UnsupportedFormatError(f'Unsupported source asset type: {mime_type}')

        if to_seconds <= 0:
            to_seconds = asset.duration + to_seconds

        duration = float(to_seconds) - float(from_seconds)

        if duration <= 0:
            raise ValueError('Start time must be before end time')

        result = io.BytesIO()
        with _FFmpegContext(asset.essence, result) as ctx:
            command = ['ffmpeg', '-v', 'error',
                       '-ss', str(float(from_seconds)), '-t', str(duration),
                       '-i', ctx.input_path, '-codec', 'copy',
                       '-f', encoder_name, '-y', ctx.output_path]

            try:
                subprocess.run(command, stderr=subprocess.PIPE, check=True)
            except subprocess.CalledProcessError as ffmpeg_error:
                error_message = ffmpeg_error.stderr.decode('utf-8')
                raise OperatorError(f'Could not trim asset: {error_message}')

        metadata = _combine_metadata(asset,
                                     'mime_type', 'width', 'height', 'video', 'audio', 'subtitle',
                                     duration=duration)

        return Asset(essence=result, **metadata)

    @operator
    def extract_frame(self, asset: Asset, mime_type: Union[MimeType, str], seconds: float = 0) -> Asset:
        """
        Creates a new image asset of the specified MIME type from the essence
        of the specified video asset.

        :param asset: Video asset which will serve as the source for the frame
        :type asset: Asset
        :param mime_type: MIME type of the destination image
        :type mime_type: MimeType or str
        :param seconds: Offset of the frame in seconds
        :type seconds: float
        :return: New image asset with converted essence
        :rtype: Asset
        """
        source_mime_type = MimeType(asset.mime_type)
        if source_mime_type.type != 'video':
            raise UnsupportedFormatError(f'Unsupported source asset type: {source_mime_type}')

        mime_type = MimeType(mime_type)
        encoder_name = self.__mime_type_to_encoder.get(mime_type)
        codec_name = self.__mime_type_to_codec.get(mime_type)
        if not (encoder_name and codec_name):
            raise UnsupportedFormatError(f'Unsupported target asset type: {mime_type}')

        result = io.BytesIO()
        with _FFmpegContext(asset.essence, result) as ctx:
            command = ['ffmpeg', '-v', 'error',
                       '-i', ctx.input_path,
                       '-ss', str(float(seconds)),
                       '-codec:v', codec_name, '-vframes', '1',
                       '-f', encoder_name, '-y', ctx.output_path]

            try:
                subprocess.run(command, stderr=subprocess.PIPE, check=True)
            except subprocess.CalledProcessError as ffmpeg_error:
                error_message = ffmpeg_error.stderr.decode('utf-8')
                raise OperatorError(f'Could not extract frame from asset: {error_message}')

        metadata = _combine_metadata(asset,
                                     'width', 'height',
                                     mime_type=mime_type)
        if 'video' in asset.metadata:
            metadata['depth'] = asset.metadata['video']['depth']

        return Asset(essence=result, **metadata)

    @operator
    def crop(self, asset: Asset, x: int, y: int, width: int, height: int) -> Asset:
        """
        Creates a cropped video asset whose essence is cropped to the specified
        rectangular area.

        :param asset: Video asset whose contents will be cropped
        :type asset: Asset
        :param x: Horizontal offset of the cropping area from left
        :type x: int
        :param y: Vertical offset of the cropping area from top
        :type y: int
        :param width: Width of the cropping area
        :type width: int
        :param height: Height of the cropping area
        :type height: int
        :return: New asset with cropped essence
        :rtype: Asset
        """
        mime_type = MimeType(asset.mime_type)
        encoder_name = self.__mime_type_to_encoder.get(mime_type)
        if not encoder_name or mime_type.type != 'video':
            raise UnsupportedFormatError(f'Unsupported source asset type: {mime_type}')

        if x == 0 and y == 0 and width == asset.width and height == asset.height:
            return asset

        max_x = max(0, min(asset.width, width + x))
        max_y = max(0, min(asset.height, height + y))
        min_x = max(0, min(asset.width, x))
        min_y = max(0, min(asset.height, y))

        if min_x == asset.width or min_y == asset.height or max_x <= min_x or max_y <= min_y:
            raise OperatorError(f'Invalid cropping area: <x={x!r}, y={y!r}, width={width!r}, height={height!r}>')

        width = max_x - min_x
        height = max_y - min_y

        result = io.BytesIO()
        with _FFmpegContext(asset.essence, result) as ctx:
            command = ['ffmpeg', '-v', 'error',
                       '-i', ctx.input_path, '-codec', 'copy',
                       '-f:v', f'crop=w={width:d}:h={height:d}:x={x:d}:y={y:d}',
                       '-f', encoder_name, '-y', ctx.output_path]

            try:
                subprocess.run(command, stderr=subprocess.PIPE, check=True)
            except subprocess.CalledProcessError as ffmpeg_error:
                error_message = ffmpeg_error.stderr.decode('utf-8')
                raise OperatorError(f'Could not crop asset: {error_message}')

        metadata = _combine_metadata(asset,
                                     'mime_type', 'duration', 'video', 'audio', 'subtitle',
                                     width=width, height=height)

        return Asset(essence=result, **metadata)

    @operator
    def rotate(self, asset: Asset, angle: float, expand: bool = False) -> Asset:
        """
        Creates an asset whose essence is rotated by the specified angle in
        degrees.

        :param asset: Asset whose contents will be rotated
        :type asset: Asset
        :param angle: Angle in degrees, counter clockwise
        :type angle: float
        :param expand: If true, changes the dimensions of the new asset so it
            can hold the entire rotated essence, otherwise the dimensions of
            the original asset will be used.
        :type expand: bool
        :return: New asset with rotated essence
        :rtype: Asset
        """
        mime_type = MimeType(asset.mime_type)
        encoder_name = self.__mime_type_to_encoder.get(mime_type)
        if not encoder_name or mime_type.type != 'video':
            raise UnsupportedFormatError(f'Unsupported source asset type: {mime_type}')

        if angle % 360.0 == 0.0:
            return asset

        angle_rad = radians(angle)
        width = asset.width
        height = asset.height

        if expand:
            if angle % 180 < 90:
                width_ = asset.width
                height_ = asset.height
                angle_rad_ = angle_rad % pi
            else:
                width_ = asset.height
                height_ = asset.width
                angle_rad_ = angle_rad % pi - pi/2
            cos_a = cos(angle_rad_)
            sin_a = sin(angle_rad_)
            width = ceil(round(width_ * cos_a + height_ * sin_a, 7))
            height = ceil(round(width_ * sin_a + height_ * cos_a, 7))

        result = io.BytesIO()
        with _FFmpegContext(asset.essence, result) as ctx:
            command = ['ffmpeg', '-v', 'error',
                       '-i', ctx.input_path, '-codec', 'copy',
                       '-f:v', f'rotate=a={angle_rad:f}:ow={width:d}:oh={height:d})',
                       '-f', encoder_name, '-y', ctx.output_path]

            try:
                subprocess.run(command, stderr=subprocess.PIPE, check=True)
            except subprocess.CalledProcessError as ffmpeg_error:
                error_message = ffmpeg_error.stderr.decode('utf-8')
                raise OperatorError(f'Could not rotate asset: {error_message}')

        metadata = _combine_metadata(asset,
                                     'mime_type', 'duration', 'video', 'audio', 'subtitle',
                                     width=width, height=height)

        return Asset(essence=result, **metadata)


class FFmpegMetadataProcessor(MetadataProcessor):
    """
    Represents a metadata processor that uses FFmpeg.
    """
    __decoder_and_stream_type_to_mime_type = {
        ('matroska,webm', 'video'): MimeType('video/x-matroska'),
        ('mov,mp4,m4a,3gp,3g2,mj2', 'video'): MimeType('video/quicktime'),
        ('avi', 'video'): MimeType('video/x-msvideo'),
        ('mpegts', 'video'): MimeType('video/mp2t'),
        ('ogg', 'video'): MimeType('video/ogg'),
        ('mp3', 'audio'): MimeType('audio/mpeg'),
        ('ogg', 'audio'): MimeType('audio/ogg'),
        ('wav', 'audio'): MimeType('audio/wav'),
    }

    __mime_type_to_encoder = {
        MimeType('video/x-matroska'): 'matroska',
        MimeType('video/quicktime'): 'mov',
        MimeType('video/x-msvideo'): 'avi',
        MimeType('video/mp2t'): 'mpegts',
        MimeType('video/ogg'): 'ogg',
        MimeType('audio/mpeg'): 'mp3',
        MimeType('audio/ogg'): 'ogg',
        MimeType('audio/wav'): 'wav',
    }

    # See https://wiki.multimedia.cx/index.php?title=FFmpeg_Metadata
    metadata_keys_by_mime_type = {
        MimeType('video/x-matroska'): bidict({}),
        MimeType('video/x-msvideo'): bidict({}),
        MimeType('video/mp2t'): bidict({}),
        MimeType('video/quicktime'): bidict({}),
        MimeType('video/ogg'): bidict({}),
        MimeType('audio/mpeg'): bidict({
            'album': 'album',                   # TALB Album
            'album_artist': 'album_artist',     # TPE2 Band/orchestra/accompaniment
            'album_sort': 'album-sort',         # TSOA Album sort order
            'artist': 'artist',                 # TPE1 Lead performer(s)/Soloist(s)
            'artist_sort': 'artist-sort',       # TSOP Performer sort order
            'bpm': 'TBPM',                      # TBPM BPM (beats per minute)
            'composer': 'composer',             # TCOM Composer
            'performer': 'performer',           # TPE3 Conductor/performer refinement
            'content_group': 'TIT1',            # TIT1 Content group description
            'copyright': 'copyright',           # TCOP (Copyright message)
            'date': 'date',                     # TDRC Recording time
            'disc': 'disc',                     # TPOS Part of a set
            'disc_subtitle': 'TSST',            # TSST Set subtitle
            'encoded_by': 'encoded_by',         # TENC Encoded by
            'encoder': 'encoder',               # TSSE Software/Hardware and settings used for encoding
            'encoding_time': 'TDEN',            # TDEN Encoding time
            'file_type': 'TFLT',                # TFLT File type
            'genre': 'genre',                   # TCON (Content type)
            'isrc': 'TSRC',                     # TSRC ISRC (international standard recording code)
            'initial_key': 'TKEY',              # TKEY Musical key in which the sound starts
            'involved_people': 'TIPL',          # TIPL Involved people list
            'language': 'language',             # TLAN Language(s)
            'length': 'TLEN',                   # TLEN Length of the audio file in milliseconds
            'lyricist': 'TEXT',                 # TEXT Lyricist/Text writer
            'lyrics': 'lyrics',                 # USLT Unsychronized lyric/text transcription
            'media_type': 'TMED',               # TMED Media type
            'mood': 'TMOO',                     # TMOO Mood
            'original_album': 'TOAL',           # TOAL Original album/movie/show title
            'original_artist': 'TOPE',          # TOPE Original artist(s)/performer(s)
            'original_date': 'TDOR',            # TDOR Original release time
            'original_filename': 'TOFN',        # TOFN Original filename
            'original_lyricist': 'TOLY',        # TOLY Original lyricist(s)/text writer(s)
            'owner': 'TOWN',                    # TOWN File owner/licensee
            'credits': 'TMCL',                  # TMCL Musician credits list
            'playlist_delay': 'TDLY',           # TDLY Playlist delay
            'produced_by': 'TPRO',              # TPRO Produced notice
            'publisher': 'publisher',           # TPUB Publisher
            'radio_station_name': 'TRSN',       # TRSN Internet radio station name
            'radio_station_owner': 'TRSO',      # TRSO Internet radio station owner
            'remixed_by': 'TP4',                # TPE4 Interpreted, remixed, or otherwise modified by
            'tagging_date': 'TDTG',             # TDTG Tagging time
            'title': 'title',                   # TIT2 Title/songname/content description
            'title_sort': 'title-sort',         # TSOT Title sort order
            'track': 'track',                   # TRCK Track number/Position in set
            'version': 'TIT3',                  # TIT3 Subtitle/Description refinement

            # Release time (TDRL) can be written, but it collides with
            # recording time (TDRC) when reading;

            # AENC, APIC, ASPI, COMM, COMR, ENCR, EQU2, ETCO, GEOB, GRID, LINK,
            # MCDI, MLLT, OWNE, PRIV, PCNT, POPM, POSS, RBUF, RVA2, RVRB, SEEK,
            # SIGN, SYLT, SYTC, UFID, USER, WCOM, WCOP, WOAF, WOAR, WOAS, WORS,
            # WPAY, WPUB, and WXXX will be written as TXXX tag
        }),
        MimeType('audio/ogg'): bidict({
            'album': 'ALBUM',                   # Collection name
            'album_artist': 'album_artist',     # Band/orchestra/accompaniment
            'artist': 'ARTIST',                 # Band or singer, composer, author, etc.
            'comment': 'comment',               # Short text description of the contents
            'composer': 'COMPOSER',             # Composer
            'contact': 'CONTACT',               # Contact information for the creators or distributors
            'copyright': 'COPYRIGHT',           # Copyright attribution
            'date': 'DATE',                     # Date the track was recorded
            'disc': 'disc',                     # Collection number
            'encoded_by': 'ENCODED-BY',         # Encoded by
            'encoder': 'ENCODER',               # Software/Hardware and settings used for encoding
            'genre': 'GENRE',                   # Short text indication of music genre
            'isrc': 'ISRC',                     # ISRC number
            'license': 'LICENSE',               # License information
            'location': 'LOCATION',             # Location where track was recorded
            'performer': 'PERFORMER',           # Artist(s) who performed the work (conductor, orchestra, etc.)
            'produced_by': 'ORGANIZATION',      # Organization producing the track (i.e. the 'record label')
            'title': 'TITLE',                   # Track/Work name
            'track': 'track',                   # Track number if part of a collection or album
            'tracks': 'TRACKTOTAL',             # Total number of track number in a collection or album
            'version': 'VERSION',               # Version of the track (e.g. remix info)
        }),
        MimeType('audio/wav'): bidict({}),
    }

    def __init__(self, config: Optional[Mapping[str, Any]] = None) -> None:
        """
        Initializes a new `FFmpegMetadataProcessor`.

        :param config: Mapping with settings.
        """
        super().__init__(config)

    @property
    def formats(self) -> Iterable[str]:
        return {'ffmetadata'}

    def read(self, file: IO) -> Mapping[str, Mapping]:
        try:
            probe_data = _probe(file)
        except subprocess.CalledProcessError:
            raise UnsupportedFormatError('Unsupported file format.')

        decoder_and_stream_type = _get_decoder_and_stream_type(probe_data)
        mime_type = self.__decoder_and_stream_type_to_mime_type.get(decoder_and_stream_type)
        if not mime_type:
            raise UnsupportedFormatError('Unsupported metadata source.')

        # Extract metadata (tags) from ffprobe information
        ffmetadata = probe_data['format'].get('tags', {})
        for stream in probe_data['streams']:
            ffmetadata.update(stream.get('tags', {}))

        # Convert FFMetadata items to metadata items
        metadata = {}
        metadata_keys = self.metadata_keys_by_mime_type[mime_type]
        for ffmetadata_key, value in ffmetadata.items():
            metadata_key = metadata_keys.inv.get(ffmetadata_key)
            if metadata_key is not None:
                metadata[metadata_key] = value

        return {'ffmetadata': metadata}

    def strip(self, file: IO) -> IO:
        try:
            probe_data = _probe(file)
        except subprocess.CalledProcessError:
            raise UnsupportedFormatError('Unsupported file format.')

        decoder_and_stream_type = _get_decoder_and_stream_type(probe_data)
        mime_type = self.__decoder_and_stream_type_to_mime_type.get(decoder_and_stream_type)
        if not mime_type:
            raise UnsupportedFormatError('Unsupported metadata source.')

        # Strip metadata
        result = io.BytesIO()
        with _FFmpegContext(file, result) as ctx:
            encoder_name = self.__mime_type_to_encoder[mime_type]
            command = ['ffmpeg', '-loglevel', 'error',
                       '-i', ctx.input_path,
                       '-map_metadata', '-1', '-codec', 'copy',
                       '-y', '-f', encoder_name, ctx.output_path]
            try:
                subprocess.run(command, stderr=subprocess.PIPE, check=True)
            except subprocess.CalledProcessError as ffmpeg_error:
                error_message = ffmpeg_error.stderr.decode('utf-8')
                raise OperatorError(f'Could not strip metadata: {error_message}')

        return result

    def combine(self, file: IO, metadata_by_type: Mapping[str, Mapping]) -> IO:
        try:
            probe_data = _probe(file)
        except subprocess.CalledProcessError:
            raise UnsupportedFormatError('Unsupported file format.')

        decoder_and_stream_type = _get_decoder_and_stream_type(probe_data)
        mime_type = self.__decoder_and_stream_type_to_mime_type.get(decoder_and_stream_type)
        if not mime_type:
            raise UnsupportedFormatError('Unsupported metadata source.')

        # Validate provided metadata
        if not metadata_by_type:
            raise ValueError('No metadata provided')
        if 'ffmetadata' not in metadata_by_type:
            raise UnsupportedFormatError(f'Invalid metadata to be combined with essence: {metadata_by_type.keys()!r}')
        if not metadata_by_type['ffmetadata']:
            raise ValueError('No metadata provided')

        # Add metadata to file
        result = io.BytesIO()
        with _FFmpegContext(file, result) as ctx:
            encoder_name = self.__mime_type_to_encoder[mime_type]
            command = ['ffmpeg', '-loglevel', 'error',
                       '-f', encoder_name, '-i', ctx.input_path]

            ffmetadata = metadata_by_type['ffmetadata']
            metadata_keys = self.metadata_keys_by_mime_type[mime_type]
            for metadata_key, value in ffmetadata.items():
                ffmetadata_key = metadata_keys.get(metadata_key)
                if ffmetadata_key is None:
                    raise ValueError(f'Unsupported metadata key: {metadata_key!r}')
                command.append('-metadata')
                command.append(f'{ffmetadata_key}={value}')

            command.extend(['-codec', 'copy',
                            '-y', '-f', encoder_name, ctx.output_path])

            try:
                subprocess.run(command, stderr=subprocess.PIPE, check=True)
            except subprocess.CalledProcessError as ffmpeg_error:
                error_message = ffmpeg_error.stderr.decode('utf-8')
                raise OperatorError(f'Could not add metadata: {error_message}')

        return result



import io
from typing import Any, Callable, IO, Iterable, Mapping, Optional, Tuple
from xml.etree import ElementTree as ET

from madam.core import Asset, Dict, MetadataProcessor, Processor, UnsupportedFormatError, operator


_INCH_TO_MM = 1 / 25.4
_PX_PER_INCH = 90
_PT_PER_INCH = 1 / 72
_FONT_SIZE_PT = 12
_X_HEIGHT = 0.7


def svg_length_to_px(length: Optional[str]) -> float:
    if length is None:
        raise ValueError()

    unit_len = 2
    if length.endswith('%'):
        unit_len = 1
    try:
        value = float(length)
        unit = 'px'
    except ValueError:
        value = float(length[:-unit_len])
        unit = length[-unit_len:]

    if unit == 'em':
        return value * _PX_PER_INCH * _FONT_SIZE_PT * _PT_PER_INCH
    elif unit == 'ex':
        return value * _PX_PER_INCH * _X_HEIGHT * _FONT_SIZE_PT * _PT_PER_INCH
    elif unit == 'px':
        return value
    elif unit == 'in':
        return value * _PX_PER_INCH
    elif unit == 'cm':
        return value * _PX_PER_INCH * _INCH_TO_MM * 10
    elif unit == 'mm':
        return value * _PX_PER_INCH * _INCH_TO_MM
    elif unit == 'pt':
        return value * _PX_PER_INCH * _PT_PER_INCH
    elif unit == 'pc':
        return value * _PX_PER_INCH * _PT_PER_INCH * 12
    elif unit == '%':
        return value
    raise ValueError()


XML_NS = dict(
    dc='http://purl.org/dc/elements/1.1/',
    rdf='http://www.w3.org/1999/02/22-rdf-syntax-ns#',
    svg='http://www.w3.org/2000/svg',
    xlink='http://www.w3.org/1999/xlink',
)


def _register_xml_namespaces() -> None:
    for prefix, uri in XML_NS.items():
        if prefix == 'svg':
            prefix = ''
        ET.register_namespace(prefix, uri)


def _parse_svg(file: IO) -> Tuple[ET.ElementTree, ET.Element]:
    _register_xml_namespaces()
    try:
        tree = ET.parse(file)
    except ET.ParseError as e:
        raise UnsupportedFormatError(f'Error while parsing XML in line {e.position[0]:d}, column {e.position[1]:d}')
    root = tree.getroot()
    if root.tag not in ('{%s}svg' % XML_NS['svg'], 'svg'):
        raise UnsupportedFormatError('XML file is not an SVG file.')
    return tree, root


def _write_svg(tree: ET.ElementTree) -> IO:
    file = io.BytesIO()
    tree.write(file, xml_declaration=False, encoding='utf-8')
    file.seek(0)
    return file


class SVGProcessor(Processor):
    """
    Represents a processor that handles *Scalable Vector Graphics* (SVG) data.
    """
    def __init__(self, config: Optional[Mapping[str, Any]] = None) -> None:
        """
        Initializes a new `SVGProcessor`.

        :param config: Mapping with settings.
        """
        super().__init__(config)

    def can_read(self, file: IO) -> bool:
        try:
            _parse_svg(file)
            return True
        except UnsupportedFormatError:
            return False

    def read(self, file: IO) -> Asset:
        _, root = _parse_svg(file)

        metadata: Dict[str, Any] = dict(mime_type='image/svg+xml')
        if 'width' in root.keys():
            metadata['width'] = svg_length_to_px(root.get('width'))
        if 'height' in root.keys():
            metadata['height'] = svg_length_to_px(root.get('height'))

        file.seek(0)
        return Asset(essence=file, **metadata)

    @staticmethod
    def __remove_xml_whitespace(elem: ET.Element) -> None:
        if elem.text:
            elem.text = elem.text.strip()
        if elem.tail:
            elem.tail = elem.tail.strip()
        for child in elem:
            SVGProcessor.__remove_xml_whitespace(child)

    @staticmethod
    def __remove_elements(root: ET.Element, qname: str, keep_func: Callable[[ET.Element], bool]) -> None:
        parents = root.findall(f'.//{qname}/..', XML_NS)
        for parent in parents:
            for elem in parent.findall(f'./{qname}', XML_NS):
                if not keep_func(elem):
                    parent.remove(elem)

    @operator
    def shrink(self, asset: Asset) -> Asset:
        """
        Shrinks the size of an SVG asset.

        :param asset: Media asset to be shrunk
        :type asset: Asset
        :return: Shrunk vector asset
        :rtype: Asset
        """
        tree, root = _parse_svg(asset.essence)

        # Minify XML
        SVGProcessor.__remove_xml_whitespace(root)
        # Remove empty texts
        SVGProcessor.__remove_elements(root, 'svg:text',
                                       lambda e: bool(
                                           e.text and
                                           e.text.strip() or
                                           list(e)
                                      ))
        # Remove all empty circles with radius 0
        SVGProcessor.__remove_elements(root, 'svg:circle',
                                       lambda e: bool(
                                           list(e) or
                                           e.get('r') != '0'
                                      ))
        # Remove all empty ellipses with x-axis or y-axis radius 0
        SVGProcessor.__remove_elements(root, 'svg:ellipse',
                                       lambda e: bool(
                                           list(e) or
                                           e.get('rx') != '0' and
                                           e.get('ry') != '0'
                                      ))
        # Remove all empty rectangles with width or height 0
        SVGProcessor.__remove_elements(root, 'svg:rect',
                                       lambda e: bool(
                                           list(e) or
                                           e.get('width') != '0' and
                                           e.get('height') != '0'
                                      ))
        # Remove all patterns with width or height 0
        SVGProcessor.__remove_elements(root, 'svg:pattern',
                                       lambda e: e.get('width') != '0' and
                                                 e.get('height') != '0')
        # Remove all images with width or height 0
        SVGProcessor.__remove_elements(root, 'svg:image',
                                       lambda e: e.get('width') != '0' and
                                                 e.get('height') != '0')
        # Remove all paths without coordinates
        SVGProcessor.__remove_elements(root, 'svg:path',
                                       lambda e: bool(e.get('d', '').strip()))
        # Remove all polygons without points
        SVGProcessor.__remove_elements(root, 'svg:polygon',
                                       lambda e: bool(e.get('points', '').strip()))
        # Remove all polylines without points
        SVGProcessor.__remove_elements(root, 'svg:polyline',
                                       lambda e: bool(e.get('points', '').strip()))
        # Remove all invisible or hidden elements
        SVGProcessor.__remove_elements(root, '*',
                                       lambda e: e.get('display') != 'none' and
                                                 e.get('visibility') != 'hidden' and
                                                 e.get('opacity') != '0')
        # Remove empty groups
        SVGProcessor.__remove_elements(root, 'svg:g',
                                       lambda e: bool(list(e)))
        # Remove all invisible or hidden elements
        SVGProcessor.__remove_elements(root, 'svg:defs',
                                       lambda e: bool(list(e)))

        essence = _write_svg(tree)

        metadata = dict(mime_type='image/svg+xml')
        for metadata_key in ('width', 'height'):
            if metadata_key in asset.metadata:
                metadata[metadata_key] = asset.metadata[metadata_key]

        return Asset(essence=essence, **metadata)


class SVGMetadataProcessor(MetadataProcessor):
    """
    Represents a metadata processor that handles Scalable Vector Graphics (SVG)
    data.

    It is assumed that the SVG XML uses UTF-8 encoding.
    """
    def __init__(self, config: Optional[Mapping[str, Any]] = None) -> None:
        """
        Initializes a new `SVGMetadataProcessor`.

        :param config: Mapping with settings.
        """
        super().__init__(config)

    @property
    def formats(self) -> Iterable[str]:
        return {'rdf'}

    def read(self, file: IO) -> Mapping[str, Mapping]:
        _, root = _parse_svg(file)
        metadata_elem = root.find('./svg:metadata', XML_NS)
        if metadata_elem is None or len(metadata_elem) == 0:
            return {'rdf': {}}
        return {'rdf': {'xml': ET.tostring(metadata_elem[0], encoding='unicode')}}

    def strip(self, file: IO) -> IO:
        tree, root = _parse_svg(file)
        metadata_elem = root.find('./svg:metadata', XML_NS)

        if metadata_elem is not None:
            root.remove(metadata_elem)

        result = _write_svg(tree)
        return result

    def combine(self, file: IO, metadata: Mapping[str, Mapping]) -> IO:
        if not metadata:
            raise ValueError('No metadata provided.')
        if 'rdf' not in metadata:
            raise UnsupportedFormatError('No RDF metadata found.')
        rdf = metadata['rdf']
        if 'xml' not in rdf:
            raise ValueError('XML string missing from RDF metadata.')

        tree, root = _parse_svg(file)
        metadata_elem = root.find('./svg:metadata', XML_NS)

        if metadata_elem is None:
            metadata_elem = ET.SubElement(root, '{%(svg)s}metadata' % XML_NS)
        metadata_elem.append(ET.fromstring(rdf['xml']))

        result = _write_svg(tree)
        return result



#!/usr/bin/env python
from setuptools import setup

import versioneer


with open('README.rst') as file:
    long_description = file.read()


setup(
    name='MADAM',
    description='Digital asset management library',
    long_description=long_description,
    url='https://github.com/eseifert/madam',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author='Michael Seifert, Erich Seifert',
    author_email='mseifert@error-reports.org, dev@erichseifert.de',
    python_requires='>=3.6',
    install_requires=['bidict', 'frozendict', 'piexif', 'pillow>=8.2.0', 'zopflipy>=1.1'],
    setup_requires=['pytest-runner', 'versioneer'],
    tests_require=['mutagen', 'piexif', 'pillow>=5.0.0', 'pytest>=3.0'],
    extras_require={
        'doc': ['sphinx >=1.3', 'sphinx_rtd_theme'],
    },
    packages=['madam'],
    platforms=['POSIX'],
    license='AGPLv3',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Topic :: Multimedia :: Graphics :: Graphics Conversion',
        'Topic :: Multimedia :: Sound/Audio :: Conversion',
        'Topic :: Multimedia :: Video :: Conversion',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    keywords='asset media processing'
)
